import unittest
from dataclasses import dataclass
from datetime import datetime

from crocodsl.field import _Field
from crocodsl.func import Hash

from .arm import Arm
from .common import NoAssignment
from .feature import Feature
from .population import Population
from .rollout import Rollout
from .value import CallType
from .variant import Variant


@dataclass
class User:
    id: str


class TestFeature(unittest.IsolatedAsyncioTestCase):
    async def test_simplest(self):
        """Test the simplest configuration."""
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        # Everyone's a member, everyone gets the same variant.
        assert await f(User("one")) == "Foo"
        assert await f(User("two")) == "Foo"
        assert await f(User("three")) == "Foo"
        assert await f(User("four")) == "Foo"

    async def test_ab(self):
        """Simple A/B gate"""
        f = Feature(
            "ab_feature",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
            ],
            rollouts=[
                Rollout(
                    Rollout.DEFAULT,
                    population=Population.DEFAULT,
                    arms=[
                        Arm("a", 0.5),
                        Arm("b", 0.5),
                    ],
                ),
            ],
        )

        # A/B are weighted 50/50. It's a coincidence that it works out
        # perfectly split here.
        # ID hashes to: 0.2812531000305875
        assert await f(User("1")) == "A"
        # ID hashes to: 0.010947509244765752
        assert await f(User("2")) == "A"
        # ID hashes to: 0.8328318262036765
        assert await f(User("3")) == "B"
        # ID hashes to: 0.937214106622131
        assert await f(User("4")) == "B"

    async def test_complex(self):
        """More complicated gate."""
        f = Feature(
            name="multi_rollout_feature_full",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    population=Population.Percent(0.2, "my_seed"),
                    arms=[
                        Arm("a", weight=0.5),
                        Arm("b", weight=0.5),
                    ],
                ),
                Rollout(
                    name="test_segment_2",
                    population=Population.Explicit(["id_1", "id_2", "id_26"]),
                    arms=[Arm("a", weight=1.0)],
                ),
            ],
        )

        # The my_seed Population hashes to: 0.9791994382873984
        # So it gets the default variant.
        assert await f(User("1")) == None  # noqa: E711

        # my_seed Population hashes to: 0.11176176296782649
        # test_segment_1 hashes to: 0.19817491497580958
        # So this gets variant `a`
        assert await f(User("MemberID")) == "A"

        # my_seed hashes to: 0.19481342807941482
        # test_segment_1 hashes to: 0.7498685567763539
        # So the user gets `b`, and *not* A as stated in the Explicit Rollout.
        assert await f(User("id_26")) == "B"

        # my_seed hashes to: 0.8810651836208253
        # So user falls through and gets `a` from the Explicit variant.
        assert await f(User("id_2")) == "A"

    async def test_custom_field(self):
        """Explicit population based on custom entity field."""
        f = Feature(
            name="custom_field",
            variants=[
                Variant("a", "A"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    population=Population.Explicit(
                        ["id_1", "id_2"], id_field=_Field("custom")
                    ),
                    arms=[Arm("a", weight=1.0)],
                ),
            ],
        )

        assert await f({"id": "id_2"}) == None  # noqa: E711
        assert await f({"custom": "id_2"}) == "A"
        assert await f({"custom": "other"}) == None  # noqa: E711
        # When the attribute returns a list, the `in` test uses ANY semantics.
        assert await f({"custom": ["other", "id_2"]}) == "A"

    async def test_custom_rollout_randomizer(self):
        """Rollout should allow a custom randomization function."""
        f = Feature(
            name="custom_randomizer",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    arms=["a", "b"],
                    randomizer=Hash(_Field("custom")),
                ),
            ],
        )

        # Hash 0.2033181243131095
        assert await f({"custom": "id_1"}, log=print) == "A"
        # Hash 0.5798682197431677
        assert await f({"custom": "id_2"}, log=print) == "B"
        # Hash 0.2249997700321562
        assert await f({"custom": "id_3"}, log=print) == "A"
        # Hash 0.33751095065789694
        assert await f({"custom": "id_4"}, log=print) == "A"
        # Hash 0.8204519266698527
        assert await f({"custom": "id_5"}, log=print) == "B"

    async def test_sticky(self):
        """Feature should respect sticky assignments."""
        mock_sticky_now = datetime(2024, 1, 2, 3, 4, 5)
        f = Feature(
            name="custom_randomizer",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    arms=["a", "b"],
                    randomizer=Hash(_Field("custom")),
                ),
            ],
        )

        def _sticky_b(feature, entity):
            return "b", "B", mock_sticky_now

        def _sticky_none(feature, entity):
            return None, None, None

        def _sticky_no_assignment(feature, entity):
            raise NoAssignment

        # Hash 0.2033181243131095
        v = await f({"custom": "id_1"}, log=print)
        assert v == "A"
        assert v.call_type == CallType.ASSIGNMENT
        v = await f({"custom": "id_1"}, log=print, sticky=_sticky_b)
        assert v == "B"
        assert v.call_type == CallType.EXPOSURE
        v = await f({"custom": "id_1"}, log=print, sticky=_sticky_none)
        assert v == None  # noqa: E711
        assert v.call_type == CallType.EXPOSURE
        v = await f({"custom": "id_1"}, log=print, sticky=_sticky_no_assignment)
        assert v == "A"
        assert v.call_type == CallType.ASSIGNMENT

    async def test_sticky_error(self):
        """Should not swallow errors thrown by the sticky assignment function."""
        f = Feature(
            "test_feature",
            variants=[Variant("foo", "Foo")],
            default_arm="foo",
        )

        def _sticky_error(feature, entity):
            raise ValueError("oops!")

        with self.assertRaises(ValueError):
            await f(User("one"), log=print, sticky=_sticky_error)



# Generated from Gram.g4 by ANTLR 4.10.1
# encoding: utf-8
import sys
from io import StringIO

from antlr4 import *

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4,
        1,
        27,
        71,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        1,
        0,
        1,
        0,
        1,
        0,
        3,
        0,
        14,
        8,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        3,
        0,
        20,
        8,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        5,
        0,
        28,
        8,
        0,
        10,
        0,
        12,
        0,
        31,
        9,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        48,
        8,
        1,
        1,
        2,
        1,
        2,
        1,
        2,
        5,
        2,
        53,
        8,
        2,
        10,
        2,
        12,
        2,
        56,
        9,
        2,
        3,
        2,
        58,
        8,
        2,
        1,
        3,
        1,
        3,
        1,
        3,
        5,
        3,
        63,
        8,
        3,
        10,
        3,
        12,
        3,
        66,
        9,
        3,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        0,
        1,
        0,
        5,
        0,
        2,
        4,
        6,
        8,
        0,
        2,
        1,
        0,
        11,
        19,
        1,
        0,
        9,
        10,
        79,
        0,
        19,
        1,
        0,
        0,
        0,
        2,
        47,
        1,
        0,
        0,
        0,
        4,
        57,
        1,
        0,
        0,
        0,
        6,
        59,
        1,
        0,
        0,
        0,
        8,
        67,
        1,
        0,
        0,
        0,
        10,
        11,
        6,
        0,
        -1,
        0,
        11,
        20,
        3,
        2,
        1,
        0,
        12,
        14,
        5,
        8,
        0,
        0,
        13,
        12,
        1,
        0,
        0,
        0,
        13,
        14,
        1,
        0,
        0,
        0,
        14,
        15,
        1,
        0,
        0,
        0,
        15,
        16,
        5,
        1,
        0,
        0,
        16,
        17,
        3,
        0,
        0,
        0,
        17,
        18,
        5,
        2,
        0,
        0,
        18,
        20,
        1,
        0,
        0,
        0,
        19,
        10,
        1,
        0,
        0,
        0,
        19,
        13,
        1,
        0,
        0,
        0,
        20,
        29,
        1,
        0,
        0,
        0,
        21,
        22,
        10,
        2,
        0,
        0,
        22,
        23,
        7,
        0,
        0,
        0,
        23,
        28,
        3,
        0,
        0,
        3,
        24,
        25,
        10,
        1,
        0,
        0,
        25,
        26,
        7,
        1,
        0,
        0,
        26,
        28,
        3,
        0,
        0,
        2,
        27,
        21,
        1,
        0,
        0,
        0,
        27,
        24,
        1,
        0,
        0,
        0,
        28,
        31,
        1,
        0,
        0,
        0,
        29,
        27,
        1,
        0,
        0,
        0,
        29,
        30,
        1,
        0,
        0,
        0,
        30,
        1,
        1,
        0,
        0,
        0,
        31,
        29,
        1,
        0,
        0,
        0,
        32,
        33,
        5,
        3,
        0,
        0,
        33,
        34,
        3,
        4,
        2,
        0,
        34,
        35,
        5,
        4,
        0,
        0,
        35,
        48,
        1,
        0,
        0,
        0,
        36,
        37,
        5,
        22,
        0,
        0,
        37,
        38,
        5,
        1,
        0,
        0,
        38,
        39,
        3,
        4,
        2,
        0,
        39,
        40,
        5,
        2,
        0,
        0,
        40,
        48,
        1,
        0,
        0,
        0,
        41,
        48,
        3,
        6,
        3,
        0,
        42,
        48,
        5,
        20,
        0,
        0,
        43,
        48,
        5,
        23,
        0,
        0,
        44,
        48,
        5,
        24,
        0,
        0,
        45,
        48,
        5,
        27,
        0,
        0,
        46,
        48,
        5,
        21,
        0,
        0,
        47,
        32,
        1,
        0,
        0,
        0,
        47,
        36,
        1,
        0,
        0,
        0,
        47,
        41,
        1,
        0,
        0,
        0,
        47,
        42,
        1,
        0,
        0,
        0,
        47,
        43,
        1,
        0,
        0,
        0,
        47,
        44,
        1,
        0,
        0,
        0,
        47,
        45,
        1,
        0,
        0,
        0,
        47,
        46,
        1,
        0,
        0,
        0,
        48,
        3,
        1,
        0,
        0,
        0,
        49,
        54,
        3,
        0,
        0,
        0,
        50,
        51,
        5,
        5,
        0,
        0,
        51,
        53,
        3,
        0,
        0,
        0,
        52,
        50,
        1,
        0,
        0,
        0,
        53,
        56,
        1,
        0,
        0,
        0,
        54,
        52,
        1,
        0,
        0,
        0,
        54,
        55,
        1,
        0,
        0,
        0,
        55,
        58,
        1,
        0,
        0,
        0,
        56,
        54,
        1,
        0,
        0,
        0,
        57,
        49,
        1,
        0,
        0,
        0,
        57,
        58,
        1,
        0,
        0,
        0,
        58,
        5,
        1,
        0,
        0,
        0,
        59,
        64,
        3,
        8,
        4,
        0,
        60,
        61,
        5,
        6,
        0,
        0,
        61,
        63,
        3,
        8,
        4,
        0,
        62,
        60,
        1,
        0,
        0,
        0,
        63,
        66,
        1,
        0,
        0,
        0,
        64,
        62,
        1,
        0,
        0,
        0,
        64,
        65,
        1,
        0,
        0,
        0,
        65,
        7,
        1,
        0,
        0,
        0,
        66,
        64,
        1,
        0,
        0,
        0,
        67,
        68,
        5,
        7,
        0,
        0,
        68,
        69,
        5,
        22,
        0,
        0,
        69,
        9,
        1,
        0,
        0,
        0,
        8,
        13,
        19,
        27,
        29,
        47,
        54,
        57,
        64,
    ]


class GramParser(Parser):
    grammarFileName = "Gram.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "'('",
        "')'",
        "'['",
        "']'",
        "','",
        "'.'",
        "'$'",
        "'Not'",
        "'And'",
        "'Or'",
        "'Eq'",
        "'Ne'",
        "'Lt'",
        "'Le'",
        "'Gt'",
        "'Ge'",
        "'In'",
        "'Has'",
        "'Matches'",
        "<INVALID>",
        "'None'",
    ]

    symbolicNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "NOT",
        "AND",
        "OR",
        "EQ",
        "NE",
        "LT",
        "LE",
        "GT",
        "GE",
        "IN",
        "HAS",
        "MATCHES",
        "BOOL",
        "NULL",
        "NAME",
        "STRING",
        "INT",
        "POS_INT",
        "SP",
        "FLOAT",
    ]

    RULE_expr = 0
    RULE_value = 1
    RULE_args = 2
    RULE_nested_attr = 3
    RULE_attr = 4

    ruleNames = ["expr", "value", "args", "nested_attr", "attr"]

    EOF = Token.EOF
    T__0 = 1
    T__1 = 2
    T__2 = 3
    T__3 = 4
    T__4 = 5
    T__5 = 6
    T__6 = 7
    NOT = 8
    AND = 9
    OR = 10
    EQ = 11
    NE = 12
    LT = 13
    LE = 14
    GT = 15
    GE = 16
    IN = 17
    HAS = 18
    MATCHES = 19
    BOOL = 20
    NULL = 21
    NAME = 22
    STRING = 23
    INT = 24
    POS_INT = 25
    SP = 26
    FLOAT = 27

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class ExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return GramParser.RULE_expr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class ParensContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ExprContext
            super().__init__(parser)
            self.inner = None  # ExprContext
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(GramParser.ExprContext, 0)

        def NOT(self):
            return self.getToken(GramParser.NOT, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParens"):
                listener.enterParens(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParens"):
                listener.exitParens(self)

    class CompareContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ExprContext
            super().__init__(parser)
            self.left = None  # ExprContext
            self.op = None  # Token
            self.right = None  # ExprContext
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(GramParser.ExprContext)
            else:
                return self.getTypedRuleContext(GramParser.ExprContext, i)

        def EQ(self):
            return self.getToken(GramParser.EQ, 0)

        def NE(self):
            return self.getToken(GramParser.NE, 0)

        def LT(self):
            return self.getToken(GramParser.LT, 0)

        def LE(self):
            return self.getToken(GramParser.LE, 0)

        def GT(self):
            return self.getToken(GramParser.GT, 0)

        def GE(self):
            return self.getToken(GramParser.GE, 0)

        def IN(self):
            return self.getToken(GramParser.IN, 0)

        def HAS(self):
            return self.getToken(GramParser.HAS, 0)

        def MATCHES(self):
            return self.getToken(GramParser.MATCHES, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCompare"):
                listener.enterCompare(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCompare"):
                listener.exitCompare(self)

    class LogicalContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ExprContext
            super().__init__(parser)
            self.left = None  # ExprContext
            self.op = None  # Token
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(GramParser.ExprContext)
            else:
                return self.getTypedRuleContext(GramParser.ExprContext, i)

        def AND(self):
            return self.getToken(GramParser.AND, 0)

        def OR(self):
            return self.getToken(GramParser.OR, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLogical"):
                listener.enterLogical(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLogical"):
                listener.exitLogical(self)

    class LiteralContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def value(self):
            return self.getTypedRuleContext(GramParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLiteral"):
                listener.enterLiteral(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLiteral"):
                listener.exitLiteral(self)

    def expr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = GramParser.ExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 0
        self.enterRecursionRule(localctx, 0, self.RULE_expr, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 19
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [
                GramParser.T__2,
                GramParser.T__6,
                GramParser.BOOL,
                GramParser.NULL,
                GramParser.NAME,
                GramParser.STRING,
                GramParser.INT,
                GramParser.FLOAT,
            ]:
                localctx = GramParser.LiteralContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 11
                self.value()
                pass
            elif token in [GramParser.T__0, GramParser.NOT]:
                localctx = GramParser.ParensContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 13
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == GramParser.NOT:
                    self.state = 12
                    self.match(GramParser.NOT)

                self.state = 15
                self.match(GramParser.T__0)
                self.state = 16
                localctx.inner = self.expr(0)
                self.state = 17
                self.match(GramParser.T__1)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 29
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 3, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 27
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 2, self._ctx)
                    if la_ == 1:
                        localctx = GramParser.CompareContext(
                            self, GramParser.ExprContext(self, _parentctx, _parentState)
                        )
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_expr
                        )
                        self.state = 21
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import \
                                FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 22
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not (
                            (
                                ((_la) & ~0x3F) == 0
                                and (
                                    (1 << _la)
                                    & (
                                        (1 << GramParser.EQ)
                                        | (1 << GramParser.NE)
                                        | (1 << GramParser.LT)
                                        | (1 << GramParser.LE)
                                        | (1 << GramParser.GT)
                                        | (1 << GramParser.GE)
                                        | (1 << GramParser.IN)
                                        | (1 << GramParser.HAS)
                                        | (1 << GramParser.MATCHES)
                                    )
                                )
                                != 0
                            )
                        ):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 23
                        localctx.right = self.expr(3)
                        pass

                    elif la_ == 2:
                        localctx = GramParser.LogicalContext(
                            self, GramParser.ExprContext(self, _parentctx, _parentState)
                        )
                        localctx.left = _prevctx
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_expr
                        )
                        self.state = 24
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import \
                                FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 1)"
                            )
                        self.state = 25
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not (_la == GramParser.AND or _la == GramParser.OR):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 26
                        localctx.left = self.expr(2)
                        pass

                self.state = 31
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 3, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ValueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return GramParser.RULE_value

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class BooleanContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def BOOL(self):
            return self.getToken(GramParser.BOOL, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBoolean"):
                listener.enterBoolean(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBoolean"):
                listener.exitBoolean(self)

    class StringContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(GramParser.STRING, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterString"):
                listener.enterString(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitString"):
                listener.exitString(self)

    class NullContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NULL(self):
            return self.getToken(GramParser.NULL, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNull"):
                listener.enterNull(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNull"):
                listener.exitNull(self)

    class ArrayContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def args(self):
            return self.getTypedRuleContext(GramParser.ArgsContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArray"):
                listener.enterArray(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArray"):
                listener.exitArray(self)

    class FunctionContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NAME(self):
            return self.getToken(GramParser.NAME, 0)

        def args(self):
            return self.getTypedRuleContext(GramParser.ArgsContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunction"):
                listener.enterFunction(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunction"):
                listener.exitFunction(self)

    class AttributeContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def nested_attr(self):
            return self.getTypedRuleContext(GramParser.Nested_attrContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAttribute"):
                listener.enterAttribute(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAttribute"):
                listener.exitAttribute(self)

    class FloatContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def FLOAT(self):
            return self.getToken(GramParser.FLOAT, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFloat"):
                listener.enterFloat(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFloat"):
                listener.exitFloat(self)

    class IntContext(ValueContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a GramParser.ValueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INT(self):
            return self.getToken(GramParser.INT, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterInt"):
                listener.enterInt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitInt"):
                listener.exitInt(self)

    def value(self):
        localctx = GramParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_value)
        try:
            self.state = 47
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [GramParser.T__2]:
                localctx = GramParser.ArrayContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 32
                self.match(GramParser.T__2)
                self.state = 33
                self.args()
                self.state = 34
                self.match(GramParser.T__3)
                pass
            elif token in [GramParser.NAME]:
                localctx = GramParser.FunctionContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 36
                self.match(GramParser.NAME)
                self.state = 37
                self.match(GramParser.T__0)
                self.state = 38
                self.args()
                self.state = 39
                self.match(GramParser.T__1)
                pass
            elif token in [GramParser.T__6]:
                localctx = GramParser.AttributeContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 41
                self.nested_attr()
                pass
            elif token in [GramParser.BOOL]:
                localctx = GramParser.BooleanContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 42
                self.match(GramParser.BOOL)
                pass
            elif token in [GramParser.STRING]:
                localctx = GramParser.StringContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 43
                self.match(GramParser.STRING)
                pass
            elif token in [GramParser.INT]:
                localctx = GramParser.IntContext(self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 44
                self.match(GramParser.INT)
                pass
            elif token in [GramParser.FLOAT]:
                localctx = GramParser.FloatContext(self, localctx)
                self.enterOuterAlt(localctx, 7)
                self.state = 45
                self.match(GramParser.FLOAT)
                pass
            elif token in [GramParser.NULL]:
                localctx = GramParser.NullContext(self, localctx)
                self.enterOuterAlt(localctx, 8)
                self.state = 46
                self.match(GramParser.NULL)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(GramParser.ExprContext)
            else:
                return self.getTypedRuleContext(GramParser.ExprContext, i)

        def getRuleIndex(self):
            return GramParser.RULE_args

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArgs"):
                listener.enterArgs(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArgs"):
                listener.exitArgs(self)

    def args(self):
        localctx = GramParser.ArgsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_args)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 57
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if ((_la) & ~0x3F) == 0 and (
                (1 << _la)
                & (
                    (1 << GramParser.T__0)
                    | (1 << GramParser.T__2)
                    | (1 << GramParser.T__6)
                    | (1 << GramParser.NOT)
                    | (1 << GramParser.BOOL)
                    | (1 << GramParser.NULL)
                    | (1 << GramParser.NAME)
                    | (1 << GramParser.STRING)
                    | (1 << GramParser.INT)
                    | (1 << GramParser.FLOAT)
                )
            ) != 0:
                self.state = 49
                self.expr(0)
                self.state = 54
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == GramParser.T__4:
                    self.state = 50
                    self.match(GramParser.T__4)
                    self.state = 51
                    self.expr(0)
                    self.state = 56
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Nested_attrContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def attr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(GramParser.AttrContext)
            else:
                return self.getTypedRuleContext(GramParser.AttrContext, i)

        def getRuleIndex(self):
            return GramParser.RULE_nested_attr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNested_attr"):
                listener.enterNested_attr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNested_attr"):
                listener.exitNested_attr(self)

    def nested_attr(self):
        localctx = GramParser.Nested_attrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_nested_attr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 59
            self.attr()
            self.state = 64
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 60
                    self.match(GramParser.T__5)
                    self.state = 61
                    self.attr()
                self.state = 66
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AttrContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(GramParser.NAME, 0)

        def getRuleIndex(self):
            return GramParser.RULE_attr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAttr"):
                listener.enterAttr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAttr"):
                listener.exitAttr(self)

    def attr(self):
        localctx = GramParser.AttrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_attr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 67
            self.match(GramParser.T__6)
            self.state = 68
            self.match(GramParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[0] = self.expr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expr_sempred(self, localctx: ExprContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 2)

        if predIndex == 1:
            return self.precpred(self._ctx, 1)



import unittest
from dataclasses import dataclass

import crocodsl.field as field
import crocodsl.func as func


@dataclass
class User:
    id: str


class TestField(unittest.TestCase):
    def test_entity_id(self):
        """Test getting the entity ID."""
        assert field.ID(User("abc")) == "abc"
        assert field.ID(User(123)) == 123

    def test_repr(self):
        """Test the representation of fields."""
        expr = func.Hash(field.ID) < 0.75
        assert repr(expr) == "Hash($id) Lt 0.75"

    def test_func(self):
        """Test fields can be used in expressions."""
        expr = field.ID == "abc"
        assert expr(User("abc")) is True
        assert expr(User(123)) is False

        expr2 = func.Hash(field.ID) < 0.75
        # Hash value is 0.7054175881782409
        assert expr2(User("abc")) is True

        expr3 = func.Hash(func.Concat("a", "b", field.ID)) > 0.75
        assert expr3(User("c")) is False

        expr4 = func.Hash(func.Concat("a", "b", field.ID)) == 0.7054175881782409
        assert expr4(User("c")) is True
        assert expr4(User("d")) is False

        expr5 = field.ID.in_(["a", "b", "c"])
        assert expr5(User("a")) is True
        assert expr5(User("b")) is True
        assert expr5(User("c")) is True
        assert expr5(User("d")) is False



import abc
import inspect
from typing import Any, Optional, Sequence

from .common import ValidationError, simple_object, NowFn, default_now


class EventLogger(abc.ABC):
    """Base class to define a logger."""

    @abc.abstractmethod
    def __call__(self, event: "_EventInstance", now: NowFn = default_now):
        """Log the given event.

        Args:
            event - instantiated event
            now - Function to call to get the current time
        """
        ...


class _EventInstance:
    """Container for event properties.

    The fields available depend on the event.
    """

    def __init__(self, name: str, **kwargs):
        self.name = name
        self.args = kwargs

    def __str__(self):
        vals = ", ".join([f"{k}={v}" for k, v in self.args.items()])
        return f"[{self.name}] {vals}"

    def __repr__(self):
        return f"<EventInstance {str(self)}>"

    def __getattr__(self, attr: str):
        try:
            return self.args[attr]
        except KeyError:
            raise AttributeError(attr)

    def __eq__(self, other):
        return self.name == repr(other)

    def asdict(self, exclude: Optional[dict] = None, compress=False) -> dict:
        """Convert object to a simple dictionary.

        Args:
            exclude - Optional dictionary of redundant fields to exclude
            compress - When serializing, if a key in the data has a `name`
            field, just use that instead of including the whole object. This
            will prevent objects from being excluded entirely, if the `exclude`
            option is also set.

        Returns:
            Dictionary of values
        """
        d: dict[str, Any] = {
            "type": self.name,
            "data": dict[str, Any](),
        }

        for k, v in self.args.items():
            vd = simple_object(v)

            if isinstance(vd, dict) and "name" in vd:
                vd = vd["name"]
            elif exclude and k in exclude and exclude[k] == vd:
                continue

            d["data"][k] = vd

        return d


class _Event:
    """An abstract event for building loggers."""

    def __init__(self, name: str, slots: Optional[Sequence[str]] = None):
        self.name = name
        slots = list(slots or [])
        slots.append("call_id")
        self.slots = tuple(slots)

    def __call__(self, log: Optional[EventLogger], **kwargs):
        if not log:
            return

        for slot in self.slots:
            if slot not in kwargs:
                raise ValidationError(
                    "Event {} missing slot value for {}".format(self.name, slot)
                )

        now_fn = kwargs.pop("now", None)
        instance = _EventInstance(self.name, **kwargs)
        # Assemble args/kwargs for logger.
        log_args: list[Any] = [instance]
        log_kwargs: dict[str, Any] = {}
        # If we have a `now_fn` passed in, see if the logger accepts it.
        # This lets us support our own advanced loggers while keeping
        # backwards-compatibility for using the `print` function.
        if now_fn:
            params = inspect.signature(log).parameters
            if "now" in params:
                log_kwargs["now"] = now_fn
            else:
                ts = now_fn()
                pfx = f"[{ts}]"
                log_args.insert(0, pfx)
        log(*log_args, **log_kwargs)

    def __repr__(self):
        return self.name


# Event definitions.
#
# The following events can be used for building Logging modules to report
# decisions made by the feature gater.


EnterGate = _Event("EnterGate", ("feature", "entity"))
"""EnterGate is fired when the top-level feature begins evaluation.

Attributes:
    feature - The top-level Feature object being evaluated.
    entity - The entity to be assigned.
"""


LeaveGate = _Event("LeaveGate", ("value",))
"""LeaveGate is fired when the top-level feature is finished evaluation.

Attributes:
    value - The value of the assigned Variant.
"""


EnterFeature = _Event("EnterFeature", ("feature", "entity"))
"""EnterFeature is fired when any feature is being evaluated.

It is fired for the top-level feature as well as nested ones.

Attributes:
    feature - The feature being evaluated
    entity - The entity to be assigned
"""


StickyAssignment = _Event(
    "StickyAssignment", ("variant", "value", "assigned", "ts", "source")
)
"""StickyAssignment is fired when the sticky assignment function is evaluated.

Attributes:
    variant - Assigned variant. None if no variant is assigned.
    value - Assigned value. This can be none, which could be a legitimate
    assigned value, or could indicate there was no value assigned yet.
    assigned - Whether or not a value was assigned.
    ts - Timestamp of the assignment.
    source - Source of the assignment ("local" if cached else "remote").
"""


ChoseVariant = _Event(
    "ChoseVariant",
    (
        "variant",
        "sticky",
    ),
)
"""ChoseVariant is fired when a Variant is selected.

Attributes:
    variant - The Variant that was selected.
    sticky - Whether the assignment should be stored permanently.
"""


LeaveFeature = _Event("LeaveFeature", ("value",))
"""LeaveFeature is fired when a feature is finished evaluating.

Fired for top-level as well as nested features.

Attributes:
    value - The vaue of the assigned Variant
"""


EnterRollout = _Event("EnterRollout", ("rollout",))
"""EnterRollout is fired when a Rollout is being evaluated.

Attributes:
    rollout - The rollout being evaluated.
"""


CheckTime = _Event(
    "CheckTime", ("entity", "time_key", "time", "after", "until", "result")
)
"""CheckTime is fired when a time range is being evaluated.

Attributes:
    entity - The entity being evaluated
    time_key - The time key of the entity
    time - The resolved time
    before - Earliest allowable time
    after - Latest allowable time
    result - Whether time was in range
"""


Randomize = _Event("Randomize", ("entity", "function", "result"))
"""Randomize is fired when a coin is being flipped.

Attributes:
    entity - The entity that's being evaluated
    function - The expression being used to flip the coin
    result - The result of flipping the coin in [0, 1]
"""


LeaveRollout = _Event("LeaveRollout", ("member",))
"""LeaveRollout is fired when a Rollout is finished evaluating.

Attributes:
    member - Boolean indicating whether entity was a member of this rollout.
"""


EnterArm = _Event("EnterArm", ("arm", "cutoff", "x"))
"""EnterArm is fired when an Arm is being evaluated.

Attributes:
    arm - The arm being evaluatd
    cutoff - The upper bound of this Arm, in [0, 1]
    x - The hash value of the entity in [0, 1]
"""


LeaveArm = _Event("LeaveArm", ("matched",))
"""LeaveArm is fired when an Arm is finished evaluating.

Attributes:
    matched - Whether the entity is a member of this Arm
"""


EvaluatePopulation = _Event("EvaluatePopulation", ("population", "entity", "member"))
"""EvaluatePopulation is fired when a Population is being evaluated.

Attributes:
    population - The Population definition
    entity - The entity being evaluated
    member - Whether the entity is a member of this Population
"""


EnterVariant = _Event("EnterVariant", ("variant",))
"""EnterVariant is fired when a Variant is being evaluated.

This Variant is the chosen Variant, but it may not be the leaf Variant.

Attributes:
    variant - The Variant definition
"""


VariantRecurse = _Event("VariantRecurse", ("inner",))
"""VariantRecurse is fired when the variant is recursing.

Attributes:
    inner - the callable being recursed into
"""


LeaveVariant = _Event("LeaveVariant", ("value",))
"""LeaveVariant is fired when a Variant is finished evaluating.

Variant might not be the leaf Variant.

Attributes:
    value - The resulting value of this Variant.
"""


EvalFunc = _Event("EvalFunc", ("f", "args", "result"))
"""EvalFunc is fired when an Expression is being evaluated.

This is fired once for every operation in the expression tree.

Attributes:
    f - The name of the function being evaluated
    args - The list of arguments
    result - The evaluation result
"""


Error = _Event("Error", ("message",))
"""Error is emitted when evaluation encounters an unexpected error.

In some cases the exception can be handled internally and is only logged for
visibility; in other cases the exception will be raised as well.

Attributes:
    message - Information about the error
"""



import unittest
from datetime import datetime, UTC

from .expr import parse
from .field import _Field
from .func import Concat, Hash, In, Literal, Ge, Gt, Le, Lt, Eq, Not, Or, And, Ne


class TestExpr(unittest.TestCase):
    def test_parse_literal(self):
        """Test parsing into simple values."""
        assert parse("None")() is None
        assert parse("1")() == 1
        assert parse("1234")() == 1234
        assert parse("1.2")() == 1.2
        assert parse("True")() is True
        assert parse("False")() is False
        assert parse("0")() == 0
        assert parse("0.123")() == 0.123
        assert parse("-100")() == -100
        assert parse("-100.001")() == -100.001
        assert parse('"hello"')() == "hello"
        assert parse(r'"say \"hi\""')() == 'say "hi"'
        assert parse(r"'say \'hi\''")() == "say 'hi'"
        assert (
            parse(
                """
            "this 'should' work"
        """
            )()
            == "this 'should' work"
        )
        assert (
            parse(
                """
            'this "should" work'
        """
            )()
            == 'this "should" work'
        )
        assert parse("[1,2,3]")() == [1, 2, 3]
        assert parse("[True, 'hi', 1, 2.5, [1,2,3], None]")() == [
            True,
            "hi",
            1,
            2.5,
            [1, 2, 3],
            None,
        ]
        assert parse(r"'hi\nwith\tescapes'")() == "hi\nwith\tescapes"

    def test_parse_field(self):
        """Test parsing field attributes."""
        assert parse("$id")(None) is None
        assert parse("$id")({}) is None
        assert parse("$id")({"id": 123}) == 123
        assert parse("$id")({"foo": 123}) is None
        assert parse("[1, 2, $id]")({"id": 3}) == [1, 2, 3]

    def test_nested_field(self):
        """Test accessing nested fields."""
        d = {"a": {"b": "c"}}
        assert parse("$a.$b")(d) == "c"

        class A:
            def foo(self, *, value):
                return {"value": value}

        assert parse("$foo.$value")(A(), context={"value": "hi"}) == "hi"
        assert (
            parse("Concat($foo.$value, ', you') Eq 'hi, you'")(
                A(), context={"value": "hi"}
            )
            is True
        )

    def test_parse_compare(self):
        """Parse comparisons."""
        assert parse("1 Eq 1").equivalent(Eq(1, 1))
        assert parse("1 Ne 2").equivalent(Not(Eq(1, 2)))
        assert parse("1 Lt 2").equivalent(Lt(1, 2))
        assert parse("1 Le 2").equivalent(Le(1, 2))
        assert parse("2 Gt 1").equivalent(Gt(2, 1))
        assert parse("2 Ge 1").equivalent(Ge(2, 1))
        assert parse("1 In [1, 2, 3]").equivalent(In(1, [1, 2, 3]))

    def test_parse_logical(self):
        """Parse logical and/or."""
        assert parse("1 Eq 1 And 2 Ne 3").equivalent(And(Eq(1, 1), Ne(2, 3)))
        assert parse("1 Eq 1 Or 2 Ne 2").equivalent(Or(Eq(1, 1), Ne(2, 2)))
        assert parse("(1 Eq 1) Or (2 Ne 2)").equivalent(Or(Eq(1, 1), Ne(2, 2)))
        assert parse("Not(1 Eq 1 Or 2 Eq 2)").equivalent(Not(Or(Eq(1, 1), Eq(2, 2))))

    def test_function(self):
        """Test calling other functions."""
        assert parse("Hash('foo')").equivalent(Hash("foo"))
        assert parse("Hash('foo')")() == 0.8845447504445093
        assert parse("Hash(Concat($prefix, ':', $id))").equivalent(
            Hash(Concat(_Field("prefix"), ":", _Field("id")))
        )
        assert (
            parse("Hash(Concat($prefix, ':', $id))")({"prefix": "pfx", "id": "123"})
            == 0.07302924453117249
        )

        orig_ts = datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC)
        now_ts = datetime(2020, 1, 3, 3, 0, 0, tzinfo=UTC)
        assert parse("TimeSince($timestamp, 'd') Gt 2")(
            {"timestamp": orig_ts}, context={"now": lambda: now_ts}
        )
        assert parse("TimeSince($timestamp, 'd') Lt 3")(
            {"timestamp": orig_ts}, context={"now": lambda: now_ts}
        )

    def test_compose(self):
        """More tests for composition of operators."""
        assert parse("Hash(Concat('test-', $id)) Lt 0.5").equivalent(
            Hash(Concat("test-", _Field("id"))) < 0.5
        )
        # "test-foo" hashes to: 0.6399548871841495
        assert parse("Hash(Concat('test-', $id)) Lt 0.5")({"id": "foo"}) is False

        assert parse("Hash($id) Lt 0.5 Or $id In ['a', 'b', 'c']").equivalent(
            Or(Hash(_Field("id")) < 0.5, _Field("id").in_(["a", "b", "c"]))
        )
        # "c" hashes to 0.5555553092364088, but is in the list
        assert parse("Hash($id) Lt 0.5 Or $id In ['a', 'b', 'c']")({"id": "c"}) is True

        assert parse(
            "(Concat($first_name, ' ', $last_name) Matches 'foo') Or ($outcomes Has 'xyz')"
        ).equivalent(
            Concat(_Field("first_name"), " ", _Field("last_name"))
            .matches("foo")
            .or_(_Field("outcomes").has("xyz"))
        )

    def test_inverse(self):
        """Test that str reps of parses are also parsable."""

        def t(exp, s):
            assert str(exp) == s
            assert exp.equivalent(parse(s))
            assert parse(s).equivalent(exp)

        t(Literal(1), "1")
        t(Literal("foo"), "'foo'")
        t(Literal("fo\no"), "'fo\\no'")
        t(Literal(r'"foo"'), """'"foo"'""")
        t(Literal(r'"fo\"o"'), r"""'"fo\\"o"'""")
        t(Literal(True), "True")
        t(Literal(False), "False")
        t(Literal([1, 2, 3]), "[1, 2, 3]")
        t(
            Literal([None, False, [1, 2, "foo", 55.123, -43.41]]),
            "[None, False, [1, 2, 'foo', 55.123, -43.41]]",
        )
        t(Ne("foo", "bar"), "Not('foo' Eq 'bar')")
        t(Gt(10, 9), "Not(10 Le 9)")
        t(Ge(10, 9), "Not(10 Lt 9)")
        t(Hash("foo"), "Hash('foo')")
        t(Hash(_Field("foo")), "Hash($foo)")
        t(
            Hash(Concat("prefix", ":", _Field("foo"))),
            "Hash(Concat('prefix', ':', $foo))",
        )
        t(Eq(_Field("id"), "foo"), "$id Eq 'foo'")
        t(
            And(_Field("id") == "foo", Hash(_Field("id")) < 0.5),
            "($id Eq 'foo') And (Hash($id) Lt 0.5)",
        )
        t(
            Concat(_Field("first_name"), " ", _Field("last_name"))
            .matches("foo")
            .or_(_Field("outcomes").has("xyz")),
            "(Concat($first_name, ' ', $last_name) Matches 'foo') Or ('xyz' In $outcomes)",
        )



from antlr4 import CommonTokenStream, InputStream, ParseTreeWalker

import crocodsl.func as func

from .field import _Field
from .gram.GramLexer import GramLexer
from .gram.GramListener import GramListener
from .gram.GramParser import GramParser


def IDENT(x):
    """A no-op to return the input of a function."""
    return x


# Escape codes are the same ones used in JSON:
# https://github.com/antlr/grammars-v4/blob/master/json/JSON.g4#L42
# Any character not listed in this map will be printed literally into the
# string. This is the desired behavior for characters like quotes and slashes
# that need to be escaped, so they are not included in this map. It's also a
# good fallback for escaping characters that don't need to be escaped.
ESCAPE = {
    "n": "\n",  # newline
    "r": "\r",  # carriage return
    "t": "\t",  # tab
    "b": "\b",  # backspace
    "f": "\f",  # form feed
}
"""A map of character values for escape codes."""


class _Args:
    """Wrapper for a function's arguments."""

    def __init__(self, *x):
        self.args = x


class ExprCompiler(GramListener):
    """Compile an Expression by walking the parse tree."""

    def __init__(self, debug=False):
        super().__init__()
        # The ops/args dicts store for each context a function to be called
        # to produce some object in the language (see `func` and `field`).
        # Values get popped as they are evaluated. (This is secretly a stack.)
        self.operators = {}
        self.arguments = {}
        self.root = None
        self.debug = debug

    def enterParens(self, ctx):
        op = func.Not if ctx.NOT() else IDENT
        self._enterNode(ctx, op, [])

    def exitParens(self, ctx):
        self._leaveNode(ctx)

    def enterLogical(self, ctx):
        op = getattr(func, ctx.op.text)
        self._enterNode(ctx, op, [])

    def exitLogical(self, ctx):
        self._leaveNode(ctx)

    def enterCompare(self, ctx):
        op = getattr(func, ctx.op.text)
        self._enterNode(ctx, op, [])

    def exitCompare(self, ctx):
        self._leaveNode(ctx)

    def enterArray(self, ctx):
        self._enterNode(ctx, lambda *x: list(x), [])

    def exitArray(self, ctx):
        self._leaveNode(ctx)

    def enterLiteral(self, ctx):
        self._enterNode(ctx, IDENT, [])

    def exitLiteral(self, ctx):
        self._leaveNode(ctx)

    def enterNull(self, ctx):
        self._enterNode(ctx, lambda: None, [])

    def exitNull(self, ctx):
        self._leaveNode(ctx)

    def enterBoolean(self, ctx):
        self._enterNode(ctx, lambda x: x == "True", [ctx.BOOL().getText()])

    def exitBoolean(self, ctx):
        self._leaveNode(ctx)

    def enterString(self, ctx):
        # Strip first and last character, which are the enclosing quotes.
        raw_str = ctx.getText()[1:-1]

        # Interpret escape sequences
        s = ""
        escaped = False
        for char in raw_str:
            if escaped:
                # Get the escaped character, falling back to the original
                # character if there isn't a predefined code.
                s += ESCAPE.get(char, char)
                escaped = False
                continue
            elif char == "\\":
                escaped = True
                continue
            else:
                s += char
                continue

        self._enterNode(ctx, IDENT, [s])

    def exitString(self, ctx):
        self._leaveNode(ctx)

    def enterInt(self, ctx):
        self._enterNode(ctx, int, [ctx.getText()])

    def exitInt(self, ctx):
        self._leaveNode(ctx)

    def enterFloat(self, ctx):
        self._enterNode(ctx, float, [ctx.getText()])

    def exitFloat(self, ctx):
        self._leaveNode(ctx)

    def enterFunction(self, ctx):
        opName = ctx.NAME().getText()
        op = getattr(func, opName)
        self._enterNode(ctx, op, [])

    def exitFunction(self, ctx):
        self._leaveNode(ctx)

    def enterAttribute(self, ctx):
        self._enterNode(
            ctx, _Field, [attr.NAME().getText() for attr in ctx.nested_attr().attr()]
        )

    def exitAttribute(self, ctx):
        self._leaveNode(ctx)

    def enterArgs(self, ctx):
        self._enterNode(ctx, _Args, [])

    def exitArgs(self, ctx):
        self._leaveNode(ctx)

    def _enterNode(self, ctx, f, args):
        self.operators[ctx] = f
        self.arguments[ctx] = args

    def _leaveNode(self, ctx):
        # Get the node's operation and arguments.
        # Normally these objects can be discarded in this step since they will
        # never be seen again, but keep them in case of debugging.
        if self.debug:
            op = self.operators[ctx]
            args = self.arguments[ctx]
        else:
            op = self.operators.pop(ctx)
            args = self.arguments.pop(ctx)

        # Evaluate!
        value = op(*args)

        # If this is a nested value, add it to the parent's arguments.
        # Otherwise this must be the root node.
        has_parent = ctx.parentCtx in self.operators
        if has_parent:
            if isinstance(value, _Args):
                self.arguments[ctx.parentCtx] += value.args
            else:
                self.arguments[ctx.parentCtx].append(value)
        else:
            # Check that there aren't any orphaned parents.
            if not self.debug and (len(self.operators) or len(self.arguments)):
                raise SyntaxError("There was an error parsing the expression")
            if isinstance(value, _Args):
                # There's no way _Args can ever be parsed as a top-level
                # context, so if this happens it's probably due to a bug in
                # the implementation of this class. (E.g., the enter method for
                # the parent context is not properly registering itself on the
                # stack, or the exit method is popping it before it's finished
                # processing its children.)
                raise RuntimeError("Unexpected top-level function args")
            elif not isinstance(value, func._Expression):
                # If the result was a literal value, wrap it.
                value = func.Literal(value)
            self.root = value


def _compile(s, debug=False):
    """Parse the symbolic expression and return the full parse tree.

    Args:
        s - A string expression
        debug - Whether to track objects for debugging later. By default the
        compiler will discard nodes in the tree after visiting them, but if
        this is True it will keep things which can be useful for debugging.

    Returns:
        ExprCompiler with the parse tree.
    """
    lexer = GramLexer(InputStream(s))
    stream = CommonTokenStream(lexer)
    parser = GramParser(stream)
    compiler = ExprCompiler(debug=debug)
    walker = ParseTreeWalker()
    walker.walk(compiler, parser.expr())
    return compiler


def parse(s, debug=False):
    """Parse a given symbolic expression into an _Expression.

    Examples:
        The syntax allows simple boolean algebra expressions. The types and
        operators are the same as the ones defined in `func`.

        The dollar sign `$` is used to signify fields of the input entity. So
        to reference `User.id`, write `$id`.

        ```
            $id Eq "my_id"
            $id In ["first_id", "second_id", "third_id"]
            ($id Eq "my_id") Or ($id In ["first_id", "second_id", "third_id"])
            Hash($id) Lt 0.5
            Hash($id) Le 0.5 And $lang Eq ['en']
            Not($lang In ['ru', 'mk', 'pl'])
        ```

    Args:
        s - A string containing the symbolic expression.

    Returns:
        _Expression object representing the parsed string.
    """
    return _compile(s, debug=debug).root



import collections.abc
import dataclasses
import json
import uuid
from datetime import date, datetime
from typing import Any, Callable

from crocodsl.common import utcnow
from .rand import getrandbits


NowFn = Callable[[], datetime]
"""Function to get the current time."""

default_now: NowFn = utcnow
"""Default function to get the current time."""


def get_uuid() -> str:
    """Get a UUID (v4).

    In tests this can be made deterministic either by patching or seeding the
    random number generator.

    Returns:
        UUIDv4 as a string
    """
    # Call into `random` explicitly so it can be seeded for tests.
    return str(uuid.UUID(int=getrandbits(128)))


def seq_id(current_id: str) -> str:
    """Get a sequential ID based on the current ID.

    This breaks the UUID formatting, but has two advantages:
      - Keeps the provenance of the original event from which
        this event was derived.
      - Can help avoid non-determinism in tests.
    provenance of events based on an original ID.

    Args:
        current_id - Base ID

    Returns:
        Unique ID based on the current ID.
    """
    base_id, _, seq = current_id.partition(":")
    if seq:
        return f"{base_id}:{int(seq) + 1}"
    return f"{base_id}:1"


def is_non_string_iterable(c: Any) -> bool:
    """Check if this is a collection (but not a string).

    Args:
        c - anything
    """
    return isinstance(c, collections.abc.Iterable) and not isinstance(c, str)


def simple_object(value: Any, with_type=False):
    """Try to simplify an arbitrary object to a simple object.

    Args:
        value - Value to simplify
        with_type - Include type information in the object

    Returns:
        Hopefully a simple JSON-type object.
    """
    if is_non_string_iterable(value):
        return [simple_object(v) for v in value]

    d = {}

    # Check if this is a dataclass instance (and not a DataClass type)
    if dataclasses.is_dataclass(value) and not isinstance(value, type):
        d = dataclasses.asdict(value)
        # If a dataclass defines a synthetic `id` using the `@property`
        # decorator, make sure it gets included in the serialization.
        if hasattr(value, "id"):
            d["id"] = value.id
    elif hasattr(value, "asdict"):
        d = value.asdict()
    elif hasattr(value, "to_dict"):
        d = value.to_dict()
    elif hasattr(value, "to_json"):
        d = value.to_json()
    else:
        d = json.loads(encode_json(value))

    if not with_type:
        return d
    return {
        "type": type(value).__name__,
        "value": d,
    }


def _json_default_encoder(obj: Any) -> str:
    """Default serializer to user with json.dumps.

    Args:
        obj - anything

    Returns:
        A simple JSON type
    """
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    return str(obj)


def encode_json(data: Any) -> str:
    """Encode object as json.

    Calls json.dumps with a good serializer.

    Args:
        data - Arbitrary object.

    Returns:
        Encoded string
    """
    return json.dumps(data, sort_keys=True, default=_json_default_encoder)


class ValidationError(Exception):
    """Error to throw when validating features."""

    pass


class InvalidConfigError(Exception):
    """Error thrown when YAML config can't be loaded."""

    pass


class MissingFeatureError(Exception):
    """Error thrown when trying to access an undefined feature."""

    pass


class NoConfig(Exception):
    """Thrown when no YAML config path is provided."""

    pass


class NoReload(Exception):
    """Thrown when no reload interval is given."""

    pass


class LoadError(Exception):
    """Thrown when reloader fails to load YAML."""

    pass


class NoAssignment(Exception):
    """Entity has not been assigned any variant of this feature.

    Throw this in the `sticky` callback to indicate the feature should be
    evaluated because there is no assignment yet.
    """

    pass


class SkipLog(Exception):
    """Raise this exception to prevent logging."""

    pass



import inspect
import sys
from typing import Any
from datetime import datetime, UTC

import mmh3

# Maximum value it's possible to represent as a 64-bit unsigned int, as a float.
MAX_UINT64_F = float(0xFFFFFFFFFFFFFFFF)


def filter_kwargs(f, kwargs):
    """Filter kwargs so that we don't raise the unexpected argument error.

    Args:
        f - function to inspect
        kwargs - args to filter

    Returns:
        Filtered dictionary of kwargs
    """
    sig = inspect.signature(f)
    # If there is a **kwargs argument, pass all the arguments.
    catch_all = any(
        v.kind is inspect.Parameter.VAR_KEYWORD and v.kind.value == 4
        for v in sig.parameters.values()
    )
    return {
        k: v
        for k, v in kwargs.items()
        if catch_all
        or (
            k in sig.parameters
            and sig.parameters[k].kind is not inspect.Parameter.POSITIONAL_ONLY
        )
    }


def dispatch(f, **kwargs):
    """Call a function with dynamic dispatch.

    Args:
        f - function to call
        **kwargs - any arguments to the function

    Returns:
        Result of calling function
    """
    fk = filter_kwargs(f, kwargs)
    return f(**fk)


def get_entity_field(entity: Any, field_name: str) -> Any:
    """Quietly get a field from an entity.

    Entity can either use dot property access or brackets as a dict.

    Args:
        entity - Entity to evaluate
        field_name - Name of field to get

    Returns:
        Value of field if found, otherwise None
    """
    result = None
    try:
        result = getattr(entity, field_name)
    except AttributeError:
        try:
            result = entity[field_name]
        except Exception:
            pass
    except Exception:
        pass
    return result


def get_entity_field_functor(obj, prop: str, **kwargs):
    """Functorized `get_entity_field`.

    Look up a property from an object. If the value is callable, the method
    will be called using dynamic dispatch with the given kwargs. Otherwise the
    value will be returned directly.
    The process will be repeated until a non-callable value is returned. An
    error is thrown if a non-callable value is not returned in a reasonable
    number of iterations.

    Args:
        obj - Object to inspect
        prop - Property to get
        **kwargs - Any named args to pass to a method

    Returns:
        Value once all callables have been resolved.

    Raises:
        `RuntimeError` if the functor doesn't produce a simple (non-callable)
        value in a certain number of iterations.
        (The specific value is determined by the global Python max recursion
        limit in `sys`.)
    """
    # Make sure the functor resolves at some point
    i = sys.getrecursionlimit()
    v = get_entity_field(obj, prop)
    while i > 0:
        i -= 1
        if not callable(v):
            return v
        v = dispatch(v, **kwargs)
    else:
        raise RuntimeError("max functor resolution depth reached")


def hash_id(s: str) -> float:
    """Compute the hash of the given ID as a float.

    This can be used instead of a PRNG to compute get the effect of random
    assignments without non-determinism from even seeded PRNGs (which would be
    non-deterministic based on call order; assignments would not necessarily
    survive server restarts, for example).

    Args:
        s - string ID representing an arbitrary entity.

    Returns:
        A float in [0, 1] that will always correspond to this entity.
    """
    # Use murmurhash3 to generate the ID. It's faster than cryptographic
    # hashes like MD5, the implementation is somewhat simpler since it
    # generates a float directly, and the output is still uniform.
    #
    # Only using the first half of the full 128-bit number; this is still
    # uniformly distributed and should divide faster. Technique taken from
    # this blog post:
    # https://www.rolando.cl/blog/2018/12/how_uniform_is_md5.html
    return mmh3.hash64(s, signed=False)[0] / MAX_UINT64_F


def utcnow() -> datetime:
    """Get current timestamp as UTC.

    Returns:
        Timezone-aware UTC timestamp.
    """
    return datetime.now(UTC)



from typing import TYPE_CHECKING, Any, Optional, Sequence, Union, cast

import alligater.events as events
import crocodsl.field as field
import crocodsl.func as func

from .arm import Arm
from .common import ValidationError, NowFn, default_now
from .population import Population, PopulationSelector

if TYPE_CHECKING:
    from . import Alligater


class Rollout:
    """A Rollout describes how to distribute variants to a population.

    Variants are distributed randomly based on a set of weighted Arms to a
    Population that can be any subset of the entire population (including the
    entire population). Membership in the population can be defined with an
    arbitrary set of rules; see `Population` for more details.
    """

    # Name for the default rollout
    DEFAULT = "default"

    # Placeholder to check if the default randomizer should be used.
    DEFAULT_RANDOMIZER = "__default__"

    def __init__(
        self,
        name: str = DEFAULT,
        population: Union[PopulationSelector, str] = Population.DEFAULT,
        arms: Optional[Sequence[Union[str, Arm]]] = None,
        randomizer: Union[str, func._Expression] = DEFAULT_RANDOMIZER,
        sticky: Optional[bool] = None,
    ):
        """Construct a new Rollout.

        Args:
            name - Name of the rollout (will use DEFAULT by default)
            population - Population to use for the rollout (by default will
            rollout to everyone using the default population).
            arms - List of Arms to use for choosing treatments. This can be
            either `Arm` objects or strings specifying a Variant name. If the
            weights are missing from the Arms they will be assigned
            automatically by splitting the available space evenly among
            all unspecified arms.
            randomizer - Expression to use to randomize treatment assignment.
            By default this randomizes using the `id` attribute of the input.
        """
        self.name = name
        self.population = self._get_population(population)
        self.arms = self._get_arms(arms)
        self.randomize = self._get_randomizer(randomizer)
        self.sticky = sticky

    def _get_population(
        self, population: Union[PopulationSelector, str]
    ) -> PopulationSelector:
        """Get the full population configuration.

        Args:
            population - Usually this should be a Population object. If it's
            not, the string `default` is understood to be the default
            population, as is the value None.

        Returns:
            Instantiated population.

        Raises:
            ValueError if the population is not a recognized value.
        """
        if population is None:
            return Population.DEFAULT

        if isinstance(population, str):
            if population.lower() == "default":
                return Population.DEFAULT
            raise ValueError("Invalid population name {}".format(population))

        # Guaranteed to be a real population at this point, not a string
        return cast(PopulationSelector, population)

    def _get_arms(self, arms: Optional[Sequence[Union[str, Arm]]]) -> list[Arm]:
        """Get a list of fully-specified arms.

        Since arms can either be a string or an Arm, and the Arm's weight is
        optional, this method will fill in the gaps.

        Specifically, it instantiates strings into Arms, and distributes weight
        evenly where it's unspecified. For example, of the following are
        given:

        ```
        arms=['foo', Arm('bar', weight=0.5)]
        ```

        We will infer the first arm to be `Arm('foo', weight=0.5)`.

        Args:
            arms - list of arms (can be partially specified) as strings or
            Arms, with or without weights.

        Returns:
            List of Arms with computed weights.

        Raises:
            ValueError if arm is neither `str` nor `Arm`.
        """
        if not arms:
            return []

        remainder = 1.0
        unknown = len(arms)
        temp_result: list[Optional[Arm]] = [None] * len(arms)

        for i, arm in enumerate(arms):
            if isinstance(arm, str):
                temp_result[i] = Arm(arm)
            elif isinstance(arm, Arm):
                temp_result[i] = arm
                weight = arm.weight
                if weight is not None:
                    remainder -= weight
                    unknown -= 1
            else:
                raise ValueError("Unexpected type for arm: {}".format(type(arm)))

        # All slots are guaranteed to be filled at this point.
        result = cast(list[Arm], temp_result)

        # Divide remaining weight evenly among the unspecified arms.
        unk_weight = remainder / float(unknown) if unknown > 0 else 0.0

        # Fill in any unspecified weights
        for arm in result:
            if arm.weight is None:
                arm.weight = unk_weight

        return result

    def _get_randomizer(
        self, randomizer: Union[str, func._Expression]
    ) -> func._Expression:
        """Get the randomization function to use.

        Args:
            A randomization function or the DEFAULT_RANDOMIZER constant.

        Returns:
            A randomization function
        """
        if isinstance(randomizer, str):
            if randomizer == self.DEFAULT_RANDOMIZER:
                return func.Hash(func.Concat(self.name, ":", field.ID))
            raise ValueError(f"Unknown randomizer {randomizer}")
        return cast(func._Expression, randomizer)

    def validate(self, variants: dict[str, Any]):
        """Ensure configuration makes sense.

        Args:
            variants - dictionary of variants available to the rollout

        Raises:
            ValidationError - if rollout is invalid for any reason
        """
        # Arms
        if not self.arms:
            raise ValidationError("Expected at least one arm")

        weights = 0.0
        for arm in self.arms:
            if arm.variant_name not in variants:
                raise ValidationError(
                    "Arm refers to unexpected variant `{}`".format(arm.variant_name)
                )
            arm.validate()
            weights += arm.weight

        if weights != 1.0:
            raise ValidationError(
                "Arm weights of rollout must sum to 1.0 (got {0:.2f})".format(weights)
            )

        # Population
        if not self.population:
            raise ValidationError("Missing population in Rollout {}".format(self.name))

        self.population.validate()

        # Special validation for the default rollout. It has to be "complete."
        # In other words, it has to assign literally everyone to some variant.
        if self.name == Rollout.DEFAULT:
            if self.population != Population.DEFAULT:
                raise ValidationError(
                    "The default Rollout must use the default Population"
                )

        if not callable(self.randomize):
            raise ValidationError("Expected randomization function to be callable")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Rollout):
            return False

        return all(
            [
                self.name == other.name,
                self.population == other.population,
                self.arms == other.arms,
                self.randomize == other.randomize,
                self.sticky == other.sticky,
            ]
        )

    def __repr__(self) -> str:
        return "<Rollout name={}>".format(self.name)

    def to_dict(self) -> dict:
        return {
            "type": "Rollout",
            "name": self.name,
            "arms": [a.to_dict() for a in self.arms],
            "population": self.population.to_dict(),
            "randomizer": str(self.randomize),
            "sticky": self.sticky,
        }

    async def __call__(
        self,
        call_id: str,
        entity: dict,
        log: Optional[events.EventLogger] = None,
        gater: Optional["Alligater"] = None,
        now: NowFn = default_now,
    ) -> Optional[str]:
        """Apply this rollout to the given entity.

        Args:
            See `Feature#__call__`.

        Returns:
            A variant name if one should be applied; otherwise None.
        """
        events.EnterRollout(log, rollout=self, call_id=call_id, now=now)

        if not await self.population(call_id, entity, log=log, gater=gater, now=now):
            events.LeaveRollout(log, member=False, call_id=call_id)
            return None

        def trace(name, args, result):
            events.EvalFunc(
                log, f=name, args=args, result=result, call_id=call_id, now=now
            )

        x = self.randomize(entity, log=trace, context={"now": now})

        events.Randomize(
            log,
            entity=entity,
            function=self.randomize,
            result=x,
            call_id=call_id,
            now=now,
        )

        cutoff = 0.0
        for arm in self.arms:
            cutoff += arm.weight
            events.EnterArm(log, arm=arm, cutoff=cutoff, x=x, call_id=call_id, now=now)
            if x <= cutoff:
                events.LeaveArm(log, matched=True, call_id=call_id, now=now)
                events.LeaveRollout(log, member=True, call_id=call_id, now=now)
                return arm.variant_name
            events.LeaveArm(log, matched=False, call_id=call_id, now=now)

        # This would only happen if something was tinkered with after validation
        raise RuntimeError("Could not find arm for entity")



import math
from enum import Enum
from typing import Generic, Optional, TypeVar
from datetime import datetime

from .events import EventLogger
from .log import DeferrableLogger
from .common import NowFn, default_now


class CallType(Enum):
    ASSIGNMENT = "assignment"
    EXPOSURE = "exposure"


T = TypeVar("T")


class Value(Generic[T]):
    """Wrapper for a value returned by a feature gate.

    The value is stored in the `value` property. It is immutable. The wrapper
    class serves as a proxy, so you can interact with the value without ever
    unwrapping it in almost all cases.

    The wrapper contains the log function. If logging was deferred from the
    gate, you can call `.log()` on the wrapper to send the log to the server.
    Logs are always idempotent; it's impossible to write logs multiple times.

    If a log is deferred and never sent, it will be garbage collected.
    """

    def __init__(
        self,
        value: T,
        variant: str,
        call_id: Optional[str] = None,
        call_type: CallType = CallType.ASSIGNMENT,
        log: Optional[EventLogger] = None,
        now: NowFn = default_now,
    ):
        self._value = value
        self._variant = variant
        self._ts = now()
        self._call_id = call_id
        self._call_type = call_type
        self._log = log

    def __del__(self):
        """Clean up unsent logs when garbage collecting."""
        if isinstance(self._log, DeferrableLogger):
            self._log.drop_log(self._call_id)

    @property
    def value(self) -> T:
        """Unwrap the internal value."""
        return self._value

    @property
    def variant(self) -> str:
        """Get the variant name."""
        return self._variant

    @property
    def ts(self) -> datetime:
        """Get the timestamp the value was created."""
        return self._ts

    @property
    def call_type(self) -> CallType:
        """Check the type of call (assignment vs exposure)."""
        return self._call_type

    def log(self, extra: Optional[dict] = None):
        """Write a log to the server.

        Only needs to be called if the log is deferred.

        Args:
            extra - Optional additional data to log.
        """
        if isinstance(self._log, DeferrableLogger) and self._call_id:
            # Note that we *don't* override the timestamp here, even though we could.
            # There is some value in keeping the original value of the timestamp, since
            # it represents when a decision was made. There will be other timestamps when
            # the log gets ingested that will better represent when the exposure happened.
            self._log.write_log(self._call_id, extra=extra)
            # If the call type wasn't already an exposure, it should be now!
            self._call_type = CallType.EXPOSURE

    # == PROXY ALL THE DUNDER METHODS TO THE INTERNAL VALUE ==

    def __eq__(self, other):
        return self._value == other

    def __gt__(self, other):
        return self._value > other

    def __lt__(self, other):
        return self._value < other

    def __ge__(self, other):
        return self._value >= other

    def __le__(self, other):
        return self._value <= other

    def __ne__(self, other):
        return self._value != other

    def __hash__(self):
        return hash(self._value)

    def __format__(self, spec):
        return self._value.format(spec)

    def __bytes__(self):
        return bytes(self._value)

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return repr(self._value)

    def __bool__(self):
        return bool(self._value)

    def __len__(self):
        return len(self._value)

    def __contains__(self, x):
        return x in self._value

    def __iter__(self):
        return iter(self._value)

    def __reversed__(self):
        return reversed(self._value)

    def __getitem__(self, k):
        return self._value[k]

    def __getattr__(self, name):
        return getattr(self._value, name)

    def __and__(self, other):
        return self._value & other

    def __xor__(self, other):
        return self._value ^ other

    def __or__(self, other):
        return self._value | other

    def __add__(self, other):
        return self._value + other

    def __mul__(self, other):
        return self._value * other

    def __sub__(self, other):
        return self._value - other

    def __truediv__(self, other):
        return self._value / other

    def __floordiv__(self, other):
        return self._value // other

    def __matmul__(self, other):
        return self._value @ other

    def __mod__(self, other):
        return self._value % other

    def __divmod__(self, other):
        return divmod(self._value, other)

    def __pow__(self, other, mod=None):
        return pow(self._value, other, mod=mod)

    def __lshift__(self, other):
        return self._value << other

    def __rshift__(self, other):
        return self._value >> other

    def __radd__(self, other):
        return other + self._value

    def __rsub__(self, other):
        return other - self._value

    def __rmul__(self, other):
        return other * self._value

    def __rmatmul__(self, other):
        return other @ self._value

    def __rtruediv__(self, other):
        return other / self._value

    def __rfloordiv__(self, other):
        return other // self._value

    def __rmod__(self, other):
        return other % self._value

    def __rdivmod__(self, other):
        return divmod(other, self._value)

    def __rpow__(self, other, mod=None):
        return pow(other, self._value, mod=mod)

    def __rlshift__(self, other):
        return other << self._value

    def __rrshift__(self, other):
        return other >> self._value

    def __rand__(self, other):
        return other & self._value

    def __ror__(self, other):
        return other | self._value

    def __rxor__(self, other):
        return other ^ self._value

    def __neg__(self):
        return -self._value

    def __pos__(self):
        return +self._value

    def __abs__(self):
        return abs(self._value)

    def __invert__(self):
        return ~self._value

    def __complex__(self):
        return complex(self._value)

    def __int__(self):
        return int(self._value)

    def __float__(self):
        return float(self._value)

    def __round__(self, ndigits=None):
        return round(self._value, ndigits=ndigits)

    def __trunc__(self):
        return math.trunc(self._value)

    def __floor__(self):
        return math.floor(self._value)

    def __ceil__(self):
        return math.ceil(self._value)



import asyncio
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Optional, Union, cast
from datetime import datetime

import alligater.events as events

from .arm import Arm
from .common import NoAssignment, ValidationError, get_uuid, default_now, NowFn
from .log import log as iolog
from .population import Population
from .rollout import Rollout
from .value import CallType, Value

if TYPE_CHECKING:
    from . import Alligater
    from .cache import AssignmentCache
    from .variant import Variant


ExistingAssignment = tuple[str, Any, datetime]
"""An existing variant name / value / ts that has been assigned."""

AsyncAssignmentFetcher = Callable[["Feature", Any], Awaitable[ExistingAssignment]]
"""Fetch assignments asynchronously."""

SyncAssignmentFetcher = Callable[["Feature", Any], ExistingAssignment]
"""Fetch assignments synchronously."""

AssignmentFetcher = Union[AsyncAssignmentFetcher, SyncAssignmentFetcher]
"""Function to fetch existing assignments for a given feature/entity."""


class Feature:
    """A Feature assigns a treatment using an arbitrary set of rules.

    The available treatments are specified as a list of Variants, which
    themselves can include other Features. (When Features are nested like this
    they are automatically evaluated to the leaf Variants.)

    Variant assignment is governed by the rules in the Rollouts. A Rollout
    specifies how to distribute treatments among a Population. The Population
    can be defined using an arbitrary set of selection criteria -- it might
    represent 100% of users, or a randomly-selected 25% of users, or only users
    who fit certain criteria. The Variant that will be assigned to a member
    of the Population is chosen randomly by the weights assigned to the
    Rollout's treatment Arms.

    Multiple Rollouts can be specified. A default Rollout comprised of the
    entire population *must* be specified, to guarantee that every input will
    be assigned a Variant (even if that Variant is something trivial, like
    `None` or "off"). The `default_arm` arg can be used to specify this easily.
    """

    def __init__(
        self,
        name: str,
        variants: Optional[list["Variant"]] = None,
        rollouts: Optional[list[Rollout]] = None,
        default_arm: Optional[Union[str, Arm]] = None,
    ):
        """Create a new feature gate.

        A default Rollout must be provided; a shortcut to do this is to pass
        a default_arm, which will create the default Rollout automatically.

        Args:
            name - The name of the feature
            variants - Full list of Variants that can be returned by this gate.
            rollouts - List of Rollouts.
            default_arm - Optional default Arm to use (only pass this if no
            default Rollout is included in the `rollouts` list). Can either be
            a string name of a variant, or an Arm object. If the weight is
            specified of the Arm, it must be `1.0`.

        Raises:
            ValidationError if the configuration isn't correct.
        """
        self.name = name
        # Store variants as a map for faster lookup
        self.variants = {v.name: v for v in variants or []}
        self.rollouts = rollouts or []

        # Create a default rollout if one was specified
        if default_arm:
            real_default_arm = (
                Arm(default_arm, weight=1.0)
                if isinstance(default_arm, str)
                else cast(Arm, default_arm)
            )

            if real_default_arm.weight is None:
                real_default_arm.weight = 1.0

            default_rollout = Rollout(
                name=Rollout.DEFAULT,
                population=Population.DEFAULT,
                arms=[real_default_arm],
            )
            self.rollouts.append(default_rollout)

        # Make sure the configuration actually makes sense.
        self.validate()

    def validate(self):
        """Ensure the feature configuration makes sense.

        Raises:
            ValidationError - if feature is invalid for any reason.
        """
        # Variants
        if not self.variants:
            raise ValidationError("At least one variant must be defined")
        [v.validate() for v in self.variants.values()]

        # Rollouts
        if not self.rollouts:
            raise ValidationError("At least one rollout needs to be defined.")

        if not any(r.name == Rollout.DEFAULT for r in self.rollouts):
            raise ValidationError(
                "At least one rollout named {} must be defined".format(Rollout.DEFAULT)
            )

        if not self.rollouts[-1].name == Rollout.DEFAULT:
            raise ValidationError(
                "The {} rollout must be the last rollout in the list".format(
                    Rollout.DEFAULT
                )
            )

        [r.validate(self.variants) for r in self.rollouts]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Feature):
            return False

        return all(
            [
                self.name == other.name,
                self.variants == other.variants,
                self.rollouts == other.rollouts,
            ]
        )

    def __repr__(self) -> str:
        return "<Feature name={}>".format(self.name)

    def to_dict(self) -> dict:
        return {
            "type": "Feature",
            "name": self.name,
            "variants": {k: v.to_dict() for k, v in self.variants.items()},
            "rollouts": [r.to_dict() for r in self.rollouts],
        }

    async def __call__(
        self,
        entity: Any,
        log: Optional[events.EventLogger] = None,
        call_id: Optional[str] = None,
        sticky: Optional[AssignmentFetcher] = None,
        assignment_cache: Optional["AssignmentCache"] = None,
        gater: Optional["Alligater"] = None,
        now: NowFn = default_now,
    ) -> Value[Any]:
        """Apply the gate to the given entity.

        Args:
            entity - the entity to gate
            log - logging function
            sticky - optional function to fetch previous assignment
            assignment_cache - optional local cache of assignments. This is
            only checked if `sticky` is also passed. It should be used to
            avoid race conditions.

        Internal Args:
            call_id - the ID of the feature invocation that this call is
                      nested within. This is None if the call is not nested.
            gater - the Alligater instance
            now - function to get the current time

        Returns:
            Variant that the entity should receive and the CallID.

            The CallID can be used for deferred logging invocations.
        """
        nested = call_id is not None
        variant_name = None
        value = None
        ts = None
        source: Optional[str] = None
        if not nested:
            call_id = get_uuid()
            events.EnterGate(log, feature=self, entity=entity, call_id=call_id, now=now)

        events.EnterFeature(log, feature=self, entity=entity, call_id=call_id, now=now)

        if sticky:
            has_assignment = False

            try:
                # Look up the assignment in the local cache first if possible.
                # This avoids race conditions that come from assignment lookups
                # that happen before they are first written to the server.
                cached = None
                if assignment_cache:
                    cached = assignment_cache.get(self, entity)

                if cached:
                    source = "local"
                    variant_name, value, ts = cached
                else:
                    if asyncio.iscoroutinefunction(sticky):
                        variant_name, value, ts = await cast(
                            AsyncAssignmentFetcher, sticky
                        )(self, entity)
                    else:
                        variant_name, value, ts = cast(SyncAssignmentFetcher, sticky)(
                            self, entity
                        )
                    source = "remote"
                has_assignment = True
            except NoAssignment:
                pass
            except Exception as e:
                events.Error(
                    log,
                    message=f"error evaluating sticky assignment {e}",
                    call_id=call_id,
                    now=now,
                )

                # Don't try to swallow exceptions, since it's now ambiguous
                # whether a value was assigned and it could be problematic for
                # an experiment to reassign something. If for the use-case it
                # doesn't matter whether the feature is re-evaluated, then
                # exceptions should be handled in the `sticky` function itself.
                raise
            finally:
                events.StickyAssignment(
                    log,
                    variant=variant_name,
                    value=value,
                    assigned=has_assignment,
                    ts=ts,
                    source=source,
                    call_id=call_id,
                    now=now,
                )
                if has_assignment:
                    events.LeaveFeature(log, value=value, call_id=call_id, now=now)
                    events.LeaveGate(log, value=value, call_id=call_id, now=now)
                    return Value(  # noqa: B012
                        value,
                        variant_name or "",
                        call_id,
                        CallType.EXPOSURE,
                        log=log,
                        now=lambda: cast(datetime, ts),
                    )

        for r in self.rollouts:
            variant_name = await r(
                cast(str, call_id), entity, log=log, gater=gater, now=now
            )
            if variant_name:
                variant = self.variants[variant_name]

                # By default, this assignment will be permanent if we have a
                # function for sticky assignments. This can be overridden at
                # the rollout level, which can specify explicitly whether or
                # not we want the assignment to be permanent.
                is_sticky_assignment = bool(sticky) if r.sticky is None else r.sticky
                if is_sticky_assignment and not sticky:
                    iolog.warning(
                        f"🏒 Rollout {r.name} requests a persistent (sticky) assignment, "
                        "but no sticky assignment fetcher was passed into Alligater. "
                        "This means assignments are probably being written but never read, "
                        "which seems like an error in your code!"
                    )

                events.ChoseVariant(
                    log,
                    variant=variant,
                    sticky=is_sticky_assignment,
                    call_id=call_id,
                    now=now,
                )
                value = await variant(call_id, entity, log=log, gater=gater, now=now)

                events.LeaveFeature(log, value=value, call_id=call_id, now=now)
                if not nested:
                    events.LeaveGate(log, value=value, call_id=call_id, now=now)

                v = Value(
                    value, variant_name, call_id, CallType.ASSIGNMENT, log=log, now=now
                )
                if assignment_cache:
                    assignment_cache.set(self, entity, variant_name, value, v.ts)
                return v

        # This code is probably unreachable since there has to be a default
        # rollout to fall back on. But the feature definitions can be quite
        # complex and I certainly don't know everything.
        events.Error(log, message="no variant found", call_id=call_id, now=now)
        events.LeaveFeature(log, value=None, call_id=call_id, now=now)
        if not nested:
            events.LeaveGate(log, value=None, call_id=call_id, now=now)

        raise RuntimeError("No variant found")



import random
from typing import Optional, Union

_rand: Union[random.Random, random.SystemRandom] = random.SystemRandom()


def seed(num: Optional[int] = None):
    """Seed the random number generator.

    By default, and if no seed is given, we will use the system's best random
    number generator available (i.e., true random). For production this is
    good; in tests you can seed to create predictable results.
    """
    global _rand
    if num is None:
        _rand = random.SystemRandom()
    _rand = random.Random(num)


def getrandbits(n: int) -> int:
    """Proxy to random.getrandbits using our RNG.

    Args:
        n - number of bits

    Returns:
        Random information with `n` bits.
    """
    global _rand
    return _rand.getrandbits(n)



import traceback

import requests
import yaml
from typing import Callable

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader  # type: ignore

from crocodsl import parse as parse_expression

from .arm import Arm
from .common import InvalidConfigError
from .feature import Feature
from .log import log
from .population import Population
from .rollout import Rollout
from .variant import Variant


ConfigSource = str | Callable[[], str]


def _expand_variants(variants):
    """Parse the variants list specified in YAML.

    Args:
        variants - Map of variants from the YAML config

    Returns:
        List of Variants expected by the Feature
    """
    result = []

    if isinstance(variants, dict):
        variants = [{"name": k, "value": v} for k, v in variants.items()]

    if isinstance(variants, list):
        for v in variants:
            value = v["value"]
            if isinstance(value, dict) and "feature" in value and len(value) == 1:
                # If the value is a nested feature, recurse
                args = _expand_feature(value["feature"])
                value = Feature(**args)
            variant = Variant(v["name"], value)
            result.append(variant)
    else:
        raise ValueError("Expected variants to be a list or map")

    return result


def _expand_expression(expression):
    """Parse an expression from string. See `expr` for more info on syntax.

    Args:
        expression - Encoded expression

    Returns:
        Decoded _Expression
    """
    return parse_expression(expression)


def _expand_population(population):
    """Instantiate a Population object from the raw YAML.

    Args:
        population - The population specification. If this is None, no
        population will be returned (which will result in the default
        population being used.

    Returns:
        Instantiated population, if any. In the case of the default population,
        the string `default` can also be used as shorthand.

    Raises:
        NotImplementedError if an unsupported population specification is provided.
    """
    if not population:
        return None

    if isinstance(population, str):
        return population

    t = population["type"].lower()
    if t == "percent":
        kwargs = {}

        if "field" in population:
            kwargs["id_field"] = _expand_expression(population["field"])

        return Population.Percent(population["value"], population["seed"], **kwargs)
    elif t == "expression":
        expr = _expand_expression(population["value"])
        return Population.Expression(expr)
    elif t == "feature":
        expr = _expand_expression(population["where"])
        return Population.Feature(population["name"], expr)
    elif t == "explicit":
        kwargs = {}

        if "field" in population:
            kwargs["id_field"] = _expand_expression(population["field"])

        return Population.Explicit(population["value"], **kwargs)
    else:
        raise NotImplementedError("Unknown population type {}".format(t))


def _expand_arms(arms):
    """Instantiate Arm objects from a list of arms.

    Args:
        arms - list of raw arms from YAML. List items should either be strings
        or dictionaries containing keys `variant` and optionally `weight`.

    Returns:
        List of arms that can be passed to a Rollout for instantiation.

    Raises:
        ValueError when an item is an unexpected type.
    """
    result = [None] * len(arms)

    for i, arm in enumerate(arms):
        if isinstance(arm, dict):
            args = {}

            if "variant" in arm:
                args["variant_name"] = arm["variant"]

            if "variant_name" in arm:
                args["variant_name"] = arm["variant_name"]

            if "weight" in arm:
                args["weight"] = arm["weight"]

            result[i] = Arm(**args)
        elif isinstance(arm, str):
            result[i] = arm
        else:
            raise ValueError("Unexpected type of arm: {}".format(type(arm)))

    return result


def _expand_rollouts(rollouts):
    """Instantiate Rollouts from the raw list.

    Args:
        rollouts - List of rollout dictionaries

    Returns:
        Instantiated Rollout list.
    """
    result = []

    for rollout in rollouts:
        args = rollout.copy()

        if "population" in args:
            args["population"] = _expand_population(args["population"])

        if "arms" in args:
            args["arms"] = _expand_arms(args["arms"])

        if "randomizer" in args:
            args["randomizer"] = _expand_expression(args["randomizer"])

        result.append(Rollout(**args))

    return result


def _expand_default_arm(default_arm):
    """Instantiate an `Arm` from the default_arm option.

    Args:
        default_arm - name of variant to use as the default treatment.

    Returns:
        Instantiated Arm with weight=1.0.
    """
    return Arm(variant_name=default_arm, weight=1.0)


def _expand_feature(feature, default_feature=None):
    """Instantiate a feature from raw YAML.

    Args:
        feature - Dictionary containing feature specification
        default_feature - optional Feature to use as default arguments.

    Returns:
        Instantiated Feature object.
    """
    result = {}
    if default_feature:
        result.update(default_feature.to_dict())

    if "name" in feature:
        if "name" in result and result["name"] != feature["name"]:
            raise ValueError(
                "Names of default feature and YAML feature differ! Got {}, expected {}".format(
                    result["name"], feature["name"]
                )
            )
        result["name"] = feature["name"]

    if "variants" in feature:
        result["variants"] = _expand_variants(feature["variants"])

    if "default_arm" in feature:
        result["default_arm"] = _expand_default_arm(feature["default_arm"])
        # If the new YAML specifies a default arm, make sure to clear out any
        # default rollout specified in the original feature.
        if "rollouts" in result:
            result["rollouts"] = [
                r for r in result["rollouts"] if r["name"] != Rollout.DEFAULT
            ]

    if "rollouts" in feature:
        result["rollouts"] = _expand_rollouts(feature["rollouts"])

    # `type` is extraneous, remove it
    if "type" in result:
        del result["type"]

    name = result.pop("name")
    return Feature(name, **result)


def _parse_feature_yaml_str(s, default_features=None):
    """Parse Features from a YAML string.

    The parsed features will override any default features supplied.

    Args:
        s - YAML string
        default_features - Optional dictionary of predefined features.

    Returns:
        Dictionary of Features.
    """
    result = {}
    if default_features:
        result.update(default_features)

    for doc in yaml.load_all(s, Loader=Loader):
        name = doc["feature"]["name"]
        default = default_features.get(name, None) if default_features else None
        feature = _expand_feature(doc["feature"], default_feature=default)
        result[name] = feature

    return result


def parse_yaml(s, default_features=None, raise_exceptions=False):
    """Load features from the given YAML config.

    The default definitions of these features can be given in
    `default_features`. The YAML lets different deployments override the
    default implementation.

    Args:
        s - YAML config as a string
        default_features - A dictionary containing the default feature
        definitions by name.
        raise_exceptions - Whether to force exceptions to be raised. Exceptions
        will otherwise be swallowed unless there are no default features.

    Returns:
        Dictionary of instantiated features.

    Raises:
        InvalidConfigError - If the config wasn't found and default_args
        was not specified.
    """
    try:
        return _parse_feature_yaml_str(s, default_features=default_features)
    except Exception as e:
        if default_features is None or raise_exceptions:
            raise InvalidConfigError(str(e)) from e
        else:
            log.warning("😖 Failed to load features from YAML: {}".format(e))
            traceback.print_exc()
            return default_features


_allowed_headers = {"authorization", "accept", "content-type"}


def load_config(source: ConfigSource, **kwargs: dict) -> str:
    """Load a string version of the yaml from the given source.

    Source can be either a local file or a remote URL, or a function.

    Args:
        source - Local file path or remote URL or a function that returns a string.

    Kwargs:
        authorization - Value to use for `Authorization` header, e.g.:
                        "Bearer my.access.token" -- only relevant for remote URLs.
        Other kwargs are passed to the function if it's a callable.

    Returns:
        String contents of the file.
    """
    if callable(source):
        return source(**kwargs)
    elif not isinstance(source, str):
        raise ValueError("Expected a string or callable")

    # Load from a remote path if given something that looks like a URL
    if source.startswith("http://") or source.startswith("https://"):
        headers: dict[str, str] = {}
        for k, v in kwargs.items():
            if k.lower() in _allowed_headers:
                headers[k] = str(v)
        r = requests.get(source, headers=headers)
        if r.status_code < 200 or r.status_code >= 300:
            raise RuntimeError(
                "Failed to load config. Got status {}".format(r.status_code)
            )
        return r.text
    else:
        # Assume it was a local file path
        with open(source) as f:
            return f.read()



import atexit
import hashlib
import threading
from functools import partial
from typing import Optional, Callable, Any, Tuple

from .arm import Arm
from .cache import AssignmentCache
from .common import (
    LoadError,
    MissingFeatureError,
    NoAssignment,
    NoConfig,
    NoReload,
    SkipLog,
    ValidationError,
    encode_json,
    simple_object,
    default_now,
    NowFn,
)
from .feature import Feature
from .events import EventLogger
from .log import (
    DeferrableLogger,
    NetworkLogger,
    ObjectLogger,
    PrintLogger,
    default_logger,
    log,
)
from .parse import load_config, parse_yaml, ConfigSource
from .population import Population
from .rand import seed
from .rollout import Rollout
from .value import CallType, Value
from .variant import Variant


StickyFn = Callable[[Feature, Any], Tuple[str, Any, int]]


class Alligater:
    """The Alligater object represents the universe of available features.

    The Alligater instance evaluates features to return an assignment. For
    example:

    ```
    gater = Alligater(yaml='path/to/my/features.yaml')

    # Assuming a feature with name "my_feature" was defined in the YAML:
    treatment = gater.my_feature(some_user)
    ```

    The Alligater instance can be configured with different modes of logging
    in order to preserve and/or debug the assignment decisions it makes.
    """

    def __init__(
        self,
        features: list[Feature] | dict[str, Feature] | None = None,
        yaml: ConfigSource | None = None,
        logger: EventLogger = default_logger,
        reload_interval: float = 0,
        sticky: StickyFn | None = None,
        loader_kwargs: dict | None = None,
        now: NowFn = default_now,
    ):
        """Create a new feature gater.

        Features can either come from a hardcoded predefined list or YAML.
        If YAML is specified, it will override the hardcoded list.

        Args:
            features - List or dictionary of features.
            yaml - Path to features specified in YAML.
            logger - Function to call to log decisions. See `events` for more
            information about how to interpret gating decisions.
            reload_interval - Number of seconds between checking for YAML changes
            sticky - Optional function to lookup previous assignments for the
            input entity. This will skip full evaluation.
            loader_kwargs - Arguments to pass to the config loader. See the
            method in `parse.py` for details.
            now - Function to call to get the current time.
        """
        log.info("🐊 Loading alligater ...")

        if isinstance(features, list):
            features = {f.name: f for f in features}

        # Current time
        self._now = now
        # Trace/event logging function
        self._logger = logger
        # Mutex for shared access to features dict from reloader thread.
        self._cv = threading.Condition()
        # Path to YAML config (either local or remote path)
        self._yaml = yaml
        # Current list of features
        self._features = features.copy() if features else {}
        # Original hard-coded list of features
        self._original_features = features.copy() if features else {}
        # Checksum of the currently loaded YAML config
        self._old_sum = ""
        # Number of seconds to wait between reloads
        self._reload_interval = reload_interval
        # Arguments for `load_config` (See `parse#load_config`.)
        self._loader_kwargs = loader_kwargs if loader_kwargs else {}
        # Sticky assignment fetcher
        self._sticky = sticky
        # Whether loader thread is stopped
        self._stopped = True
        # Background thread
        self._thread = None
        self._local_assignments = AssignmentCache()

        # Start reloading. This will load one initial time on the main thread,
        # then (if `reload_interval` and `yaml` options are passed) will reload
        # at interval forever in a background thread.
        try:
            self._reload()
        except NoConfig:
            log.warning("😭 No feature path specified for Alligater.")
        except NoReload:
            log.warning("😤 No reload interval specified for Alligater.")

        if not self._features:
            log.warning(
                "🤬 Alligater was instantiated without any features! Was this intentional?"
            )

    def __getattr__(self, feature_name):
        """Get a function that will evaluate the given feature.

        Args:
            feature_name - Name of feature to evaluate

        Returns:
            Function that can be called with an `entity` to evaluate.
        """
        try:
            with self._cv:
                ft = self._features[feature_name]
            return partial(self, ft)
        except KeyError as e:
            raise MissingFeatureError(feature_name) from e

    async def __call__(
        self, feature, entity, silent=False, deferred=None, now: Optional[NowFn] = None
    ):
        """Evaluate an entity against the given feature.

        Args:
            feature - Feature to evaluate
            entity - Arbitrary entity to evaluate
            silent - Whether to suppress logging for this invocation
            deferred - Whether exposure logging should be deferred. Note that
            assignment logging can never be deferred.
            now - Optional function to call to get the current time. (Overrides default.)

        Returns:
            Wrapped value of the variant to return.
        """
        logger = self._logger if not silent else None
        now_func = now if now else self._now
        value = await feature(
            entity,
            log=logger,
            sticky=self._sticky,
            assignment_cache=self._local_assignments,
            gater=self,
            now=now_func,
        )
        # Note that Logger implementations that aren't DeferrableLoggers
        # log immediately and calling `log` is just a no-op.
        if value.call_type == CallType.ASSIGNMENT:
            value.log()

        # Log an exposure immediately unless it's deferred.
        if not deferred:
            value.log()

        return value

    def stop(self):
        """Stop the background reloader."""
        # Make sure that the logger stops if it can.
        if self._logger and hasattr(self._logger, "stop"):
            self._logger.stop()

        # Stop internal loader thread if necessary.
        if self._stopped or not self._thread:
            return

        with self._cv:
            self._stopped = True
            self._cv.notify_all()
        self._thread.join()
        self._thread = None

    def _reload(self):
        """Start a background process to reload YAML at an interval."""
        if not self._stopped or self._thread:
            log.warning("👀 Called _reload when Alligater was already reloading")
            return

        # Force a synchronous run of the loader on the main thread
        # If this fails, it'll throw an error on the main thread.
        self._run_reload(once=True)

        # Start a background loader to reload at interval.
        if not self._reload_interval:
            raise NoReload("No reload interval passed; reloader will not run")
        thread = threading.Thread(target=self._run_reload, args=())
        self._stopped = False
        self._thread = thread
        thread.start()
        atexit.register(self.stop)

    def _run_reload(self, once=False):
        """Background worker to reload features from YAML.

        Args:
            once - Whether to force the loader to run only once, immediately.
        """
        while True:
            with self._cv:
                # Don't sleep if this is a one-time run.
                if not once:
                    self._cv.wait(timeout=self._reload_interval)
                    if self._stopped:
                        return

                if not self._yaml:
                    raise NoConfig("No YAML path is specified; (re)loader is exiting.")

                try:
                    log.debug("🕵️‍♀️  Checking for new features ...")
                    yaml_str = load_config(self._yaml, **self._loader_kwargs)
                    new_sum = self._checksum(yaml_str)
                    if new_sum != self._old_sum:
                        log.debug("🎉 New features found!")
                        # NOTE: When features are reloaded, they are **not**
                        # merged with the _current_ list of features, they are
                        # always merged with the original hardcoded list. This
                        # results in more predictable behavior -- if we always
                        # merged with the current list, transient mistakes in
                        # a config could propagate forever.
                        self._features = parse_yaml(
                            yaml_str,
                            default_features=self._original_features,
                            raise_exceptions=once,
                        )
                        self._old_sum = new_sum
                        log.debug("☺️  Updated to {}!".format(new_sum))
                        if once:
                            return
                    else:
                        log.debug("😔 No new features found")
                except Exception as e:
                    if once:
                        raise LoadError("Failed to load feature spec") from e
                    else:
                        log.error(
                            "😫 Alligater loader encountered an error: {}".format(e)
                        )

    def _checksum(self, s):
        """Compute checksum of a string.

        Args:
            s - String to hash

        Returns:
            Unique hash of the input
        """
        return hashlib.sha256(s.encode("utf-8")).hexdigest()


__all__ = [
    "Alligater",
    "Feature",
    "Variant",
    "Arm",
    "Rollout",
    "Population",
    "ValidationError",
    "events",
    "seed",
    "log",
    "encode_json",
    "simple_object",
    "DeferrableLogger",
    "NetworkLogger",
    "ObjectLogger",
    "PrintLogger",
    "NoAssignment",
    "SkipLog",
    "CallType",
    "Value",
]



from .common import ValidationError


class Arm:
    """An Arm represents a treatment group as a proportion of a population.

    The arm specifies a weight, which is a % of its parent population, and the
    name of a variant, which is the treatment this group will receive.
    """

    def __init__(self, variant_name, weight=None):
        """Create a new Arm representing the given variant.

        The weight does not need to be specified initially, but it will need
        to be set before `validate` is called.

        Args:
            variant_name - name of Variant that this Arm represents.
            weight - Optional float weight in [0, 1]
        """
        self.variant_name = variant_name
        self.weight = weight

    def __eq__(self, other):
        if not isinstance(other, Arm):
            return False

        return all(
            [
                self.variant_name == other.variant_name,
                self.weight == other.weight,
            ]
        )

    def __repr__(self):
        return "<Arm name={} weight={}>".format(self.variant_name, self.weight)

    def to_dict(self):
        return {
            "type": "Arm",
            "variant": self.variant_name,
            "weight": self.weight,
        }

    def validate(self):
        """Ensure configuration makes sense.

        Raises:
            ValidationError - if config doesn't make sense
        """
        if self.weight < 0.0:
            raise ValidationError("Arm weight must be positive")
        if self.weight > 1.0:
            raise ValidationError("Arm weight can't be greater than 1.0")



import asyncio
from typing import TYPE_CHECKING, Optional

import alligater.events as events

from .common import ValidationError, NowFn, default_now
from .feature import Feature
from .value import Value

if TYPE_CHECKING:
    from . import Alligater


class Variant:
    """A Variant represents a treatment that can be assigned.

    The Variant can store any value whatsoever, identified by a `name` that
    must be a string. This name will be used by `Arm`s to indicate which
    variant to assign.

    The value of a Variant can be another Feature. This creates a
    tree that will be evaluated recursively when the top-level feature gate
    is called.

    The same recursive evaluation can be achieved for other `callable` values
    by setting the `functor` argument to `True`.
    """

    def __init__(self, name, value, functor=False):
        """Create a container for a value to return from the feature gate.

        The value is given a `name` for convenience in logging and debugging.

        If the value is given as a Feature, the Variant will not return the
        value directly, but rather evaluate the Feature recursively. This
        allows for the creation of nested gates. This same behavior can be
        achieved for other callable values by setting `functor=True`.

        Args:
            name - The human-readable label for the variant
            value - Any value to return when the variant is applied
            functor - Whether to treat the `value` as a function that returns
            the true value. (By default this is False; the value will be
            returned literally.) If this is True, the `value` will be called
            with the arguments passed to the Feature call.

        Raises:
            ValidationError - If there's any issue with the configuration
        """
        self.name = name
        self._value = value
        self.is_nested = functor or isinstance(value, Feature)

    def validate(self):
        """Ensure the configuration of this Variant makes sense.

        Raises:
            ValidationError - If there's any issue with the configuration
        """
        if not self.name:
            raise ValidationError("Variant must be given a name")

        if self.is_nested and not callable(self._value):
            raise ValidationError("Expected the value to be callable")

    def __eq__(self, other):
        if not isinstance(other, Variant):
            return False

        return all(
            [
                self.name == other.name,
                self._value == other._value,
                self.is_nested == other.is_nested,
            ]
        )

    def __repr__(self):
        return "<Variant name={}>".format(self.name)

    def to_dict(self):
        v = self._value
        return {
            "type": "Variant",
            "name": self.name,
            "value": v.to_dict() if hasattr(v, "to_dict") else v,
            "nested": self.is_nested,
        }

    async def __call__(
        self,
        call_id,
        entity,
        log=None,
        gater: Optional["Alligater"] = None,
        now: NowFn = default_now,
    ):
        """Get the value for this variant.

        If the value is a Feature (or a `functor`), it will be called with
        all the arguments passed into this function and the result will be
        returned.

        Args:
            See `Feature#__call__`.

        Returns:
            Could be anything.
        """
        events.EnterVariant(log, variant=self, call_id=call_id, now=now)

        if self.is_nested:
            events.VariantRecurse(log, inner=self._value, call_id=call_id, now=now)

            result = None
            if asyncio.iscoroutinefunction(self._value):
                result = await self._value(
                    entity, log=log, call_id=call_id, gater=gater, now=now
                )
            else:
                result = self._value(
                    entity, log=log, call_id=call_id, gater=gater, now=now
                )

            # Unwrap wrapped Values. This happens when features are nested in
            # variants. Only the final value will be wrapped.
            # TODO(jnu): this loses the assignment TS. Unclear if that's expected behavior,
            # but it's something to keep in mind.
            if isinstance(result, Value):
                return result.value
            else:
                return result

        events.LeaveVariant(log, value=self._value, call_id=call_id, now=now)
        return self._value



import unittest
from dataclasses import dataclass

from .common import simple_object


class TestCommon(unittest.TestCase):
    def test_simple_object(self):
        @dataclass
        class Foo:
            id: str
            bar: int

        assert simple_object(Foo(id="a", bar=1)) == {
            "id": "a",
            "bar": 1,
        }
        assert simple_object(Foo(id="a", bar=1), with_type=True) == {
            "type": "Foo",
            "value": {
                "id": "a",
                "bar": 1,
            },
        }

        @dataclass
        class Bar:
            foo: str
            zap: int

            @property
            def id(self):
                return f"{self.foo}-{self.zap}"

        assert simple_object(Bar(foo="a", zap=1)) == {
            "foo": "a",
            "zap": 1,
            "id": "a-1",
        }
        assert simple_object(Bar(foo="a", zap=1), with_type=True) == {
            "type": "Bar",
            "value": {
                "foo": "a",
                "zap": 1,
                "id": "a-1",
            },
        }



import abc
import atexit
import copy
import logging
import signal
import threading
from typing import Optional

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3.util import Retry

import alligater.events as events

from .common import SkipLog, encode_json, seq_id, simple_object, default_now, NowFn

# Sys log (different than feature trace log)
log = logging.getLogger("alligater")


class DeferrableLogger(events.EventLogger):
    """Traits for a logger that can be deferred."""

    @abc.abstractmethod
    def write_log(self, call_id: str, extra: Optional[dict] = None):
        """Write the log with the given ID.

        Args:
            call_id - The UUID of the log to write
            extra - Extra data to send with this log
        """
        ...

    @abc.abstractmethod
    def drop_log(self, call_id: str):
        """Clear the log with the given ID without writing.

        This must be called to avoid memory leaks for dropped logs.
        """
        ...


class ObjectLogger(DeferrableLogger):
    """Logger that aggregates event data as an object."""

    def __init__(
        self, write, trace: bool = False, workers: int = 1, install_signals: bool = True
    ):
        """Create a new ObjectLogger with the given `write` callback.

        Args:
            write - Callback when event data is ready.
            trace - Whether to enable trace logging
            workers - Number of background workers to handle IO
            install_signals - Whether to install signal handlers for cleanup.
                If you are running in an environment like uvicorn, there may be
                conflicts with signal handlers.
        """
        self._cv = threading.Condition()
        self._write = write
        self._cache = dict[str, dict]()
        self._finished = list[dict]()
        self._deferred = set[str]()
        self._trace = trace
        self._stopped = False
        self._workers = [
            threading.Thread(
                name=f"ObjectLogger-io-{w}", target=self._write_results, daemon=True
            )
            for w in range(workers)
        ]
        for w in self._workers:
            w.start()

        # Cleanup threads and try to drain the queue if possible when exiting.
        atexit.register(self._drain)
        if install_signals:
            sigs = [signal.SIGINT, signal.SIGTERM]
            for sig in sigs:
                signal.signal(sig, self._drain)

    def stop(self):
        """Shut off the logger and send any pending messages."""
        self._drain()

    def __call__(self, event, now: NowFn = default_now):
        """Log a single event."""
        # Every event tracks an ID that is unique to the invocation.
        call_id = event.call_id

        with self._cv:
            if event == events.EnterGate:
                self._cache[call_id] = {
                    "ts": now(),
                    "call_id": call_id,
                    "entity": simple_object(event.entity, with_type=True),
                    "feature": event.feature.to_dict(),
                    "assignment": None,
                    "variant": "",
                    "trace": None if not self._trace else [],
                    "repeat": False,
                    "sticky": False,
                }

            if self._trace:
                cur = self._cache[call_id]
                d = event.asdict(
                    exclude={
                        "call_id": cur["call_id"],
                        "entity": cur["entity"]["value"],
                        "feature": cur["feature"],
                    },
                    compress=True,
                )
                self._cache[call_id]["trace"].append(d)

            if event == events.ChoseVariant:
                # There might be nested variants that get chosen if the feature
                # is defined as a tree. In that case this will be called
                # multiple times, but only the last (leaf) variant will stick.
                self._cache[call_id]["variant"] = simple_object(event.variant)
                self._cache[call_id]["sticky"] = event.sticky

            # Get assigned variant from the sticky assignment if it exists.
            if event == events.StickyAssignment and event.assigned:
                self._cache[call_id].update(
                    {
                        "variant": {
                            "name": event.variant,
                        },
                        "repeat": True,
                        "sticky": True,
                    }
                )

            if event == events.LeaveGate:
                self._cache[call_id]["assignment"] = event.value
                # Note that log is not written immediately. Call `write_log` to
                # put it in the queue.
                self._deferred.add(call_id)

    def write_log(self, call_id, extra=None):
        """Write a deferred log by its ID."""
        with self._cv:
            if call_id not in self._deferred:
                return

            # Make a copy of the event data to write. This freezes the data at
            # this moment in time.
            data = copy.deepcopy(self._cache[call_id])
            # If this event has been sent before, generate a new ID for it.
            if data["repeat"]:
                data["call_id"] = seq_id(data["call_id"])

            # Add extra data if it's given during this call.
            if extra:
                data["extra"] = extra

            # Add to queue to publish.
            self._finished.append(data)
            # Mark the cached object as a repeat if it's logged again.
            self._cache[call_id]["repeat"] = True
            # Wake a sender thread to broadcast the log.
            self._cv.notify()

    def drop_log(self, call_id):
        """Drop a deferred log by its ID."""
        with self._cv:
            self._deferred.remove(call_id)
            try:
                del self._cache[call_id]
            except KeyError:
                pass

    def _drain(self, *args):
        """Stop worker threads and drain the queue."""
        if not self._stopped:
            log.debug("🙉 Stopping log write workers ...")
            with self._cv:
                self._stopped = True
                self._cv.notify_all()
            [w.join() for w in self._workers]
            self._workers = []

        with self._cv:
            if not self._finished:
                return

            log.debug("🪵 Draining pending logs ...")
            while len(self._finished) > 0:
                event = self._finished.pop(0)
                try:
                    self._write_handled(event)
                except SystemExit:
                    continue

    def _write_results(self):
        """[THREAD] Loop over queue and write events as they are found."""
        while not self._stopped:
            with self._cv:
                if not self._finished:
                    self._cv.wait()

                if self._stopped:
                    return

                event = self._finished.pop(0)
            # Release the lock to write event, so that other workers can write.
            self._write_handled(event)

    def _write_handled(self, event):
        """Write an event with error handling."""
        try:
            self._write(event)
        except Exception as e:
            log.error("😓 Failed to write event: {}".format(e))


class NetworkLogger(ObjectLogger):
    """Logger that writes results over the network with retries."""

    def __init__(
        self,
        url,
        headers=None,
        auth=None,
        body=None,
        timeout=10.0,
        max_retries=5,
        backoff_factor=1.0,
        debug=False,
        **kwargs,
    ):
        """Create a logger that sends data to a remote endpoint.

        Args:
            url - Endpoint to log to
            headers - Key/value hash of headers to include in every request.
            This may include `Authorization` and `Content-Type` headers. The
            default content type is `application/json; charset=utf-8`.
            auth - Optional basic auth tuple
            body - Function to serialize the input log object to a string. By
            default this just does a standard JSON serialization.
            timeout - Time to wait for server response before giving up. The
            default is 10s.
            max_retries - Number of times to reattempt requests, default is 3.
            backoff_factor - Exponential backoff factor; see requests docs.
            debug - Whether to print log lines to stderr in addition to sending
            them over the network.
            **kwargs - See ObjectLogger.__init__
        """
        self._url = url
        self._debug = debug
        self._body = body
        self._timeout = timeout

        # Configure http adapter
        http = requests.Session()
        adapter = HTTPAdapter(
            max_retries=Retry(
                total=max_retries,
                status_forcelist=[413, 429, 500, 502, 503, 504],
                allowed_methods=[
                    "HEAD",
                    "OPTIONS",
                    "POST",
                    "GET",
                    "PUT",
                    "TRACE",
                    "DELETE",
                ],
                backoff_factor=backoff_factor,
            )
        )
        http.mount("http://", adapter)
        http.mount("https://", adapter)
        http.auth = auth
        http.headers.update(
            {
                "Content-Type": "application/json; charset=utf-8",
            }
        )
        if headers:
            http.headers.update(headers)
        self._http = http

        # Configure the base ObjectLogger.
        super().__init__(self._post, **kwargs)

    def _serialize(self, data):
        """Convert data to string.

        If the `body` function passed on init does not convert the data to a
        string itself, the default JSON encoder will stringify the output of
        the body function.

        Args:
            data - Event data

        Returns:
            Serialized string of data to send over network.
        """
        if self._body:
            data = self._body(data)

        if not isinstance(data, str):
            data = encode_json(data)

        return data

    def _post(self, data):
        """Write data over the network.

        Args:
            data - Log to write
        """
        self._debugw("Writing log: {}", data)

        try:
            s = self._serialize(data)
            r = self._http.post(self._url, timeout=self._timeout, data=s)

            r.raise_for_status()
            self._debugw("Log written")
        except HTTPError as e:
            self._debugw("Failed to write log: {}", e)
        except SkipLog:
            self._debugw("Skipping writing log (intentionally)")
        except Exception as e:
            self._debugw("Something unexpected happened: {}", e)

    def _debugw(self, *args):
        """Write a debug line.

        This is a no-op if debug mode is not enabled.

        Args:
            *args - anything
        """
        if not self._debug:
            return

        log.debug("🪵  " + args[0].format(*args[1:]))


class PrintLogger(events.EventLogger):
    """Logger that dumps events and features from feature evaluation."""

    def __init__(self, trace: bool = False):
        """Create a new print logger.

        Args:
            trace - Whether to dump full trace of events in decision.
        """
        self._trace = trace

    def __call__(self, event, now: NowFn = default_now):
        trace = self._trace

        if event == events.EnterGate:
            print("=== MAKING ASSIGNMENT ===")
            print("CallId:", event.call_id)
            print("Entity:", event.entity)
            print("Feature:", event.feature.to_dict())
            print("Current time:", now())
            if trace:
                print("Trace:")

        if trace:
            print(str(event))

        if event == events.ChoseVariant:
            print("Variant:", event.variant)
            print("Sticky:", event.sticky)

        if event == events.LeaveGate:
            print("Assignment:", event.value)
            print("=== END ASSIGNMENT ===")


default_logger = PrintLogger(trace=True)
"""Default Alligater logging function."""



from threading import Lock
from typing import TYPE_CHECKING, Any, Optional, Tuple
from datetime import datetime

if TYPE_CHECKING:
    from .feature import Feature


EntityId = Tuple[str, str]
"""Composite ID of an entity: class name and ID attribute."""

CachedAssignment = Tuple[str, Any, datetime]
"""Cached variant name and value and assignment time."""

AssignmentsCache = dict[str, dict[EntityId, CachedAssignment]]
"""Dictionary containing assignments."""


def _entity_id(entity: Any) -> EntityId:
    """Get the ID from an entity.

    Args:
        entity - Anything with an ID

    Returns:
        Tuple with typename and ID.
    """
    entity_id = None
    if hasattr(entity, "id"):
        entity_id = entity.id
    elif "id" in entity:
        entity_id = entity["id"]
    else:
        entity_id = repr(entity_id)

    return type(entity).__name__, entity_id


class AssignmentCache:
    """Cache feature assignments.

    Currently cache is unbounded since there is the life of the service is not
    long enough for it to matter, but could add TTL.
    """

    def __init__(self):
        self.cache = AssignmentsCache()
        self.lock = Lock()

    def clear(self):
        """Clear the cache."""
        with self.lock:
            self.cache.clear()

    def set(
        self, feature: "Feature", entity: Any, variant: str, value: Any, ts: datetime
    ):
        """Add a new item to the cache.

        Args:
            feature - Feature to look up
            entity - Entity to look up
            variant - Name of variant to set
            value - Value of the variant to set
            ts - Timestamp of the assignment
        """
        with self.lock:
            if feature.name not in self.cache:
                self.cache[feature.name] = {}
            self.cache[feature.name][_entity_id(entity)] = (variant, value, ts)

    def get(self, feature: "Feature", entity: Any) -> Optional[CachedAssignment]:
        """Look up cached assignment.

        Args:
            feature - Feature to look up
            entity - Entity to look up

        Returns:
            Tuple of cached variant name and value and ts, if it exists, or N
        """
        with self.lock:
            return self.cache.get(feature.name, {}).get(_entity_id(entity), None)



from .common import get_entity_field_functor
from .func import _Expression


class _Field(_Expression):
    """Get an attribute from the entity."""

    def __init__(self, *names):
        self.names = names

    def __call__(self, *args, log=None, context=None):
        entity = args[0]
        result = entity
        context = context or {}
        for name in self.names:
            result = get_entity_field_functor(result, name, **context)

        self._trace(log, [entity, repr(self)], result)

        return result

    def __repr__(self):
        return ".".join([f"${name}" for name in self.names])


# Field definitions
#
# These are predefined values that describe how to interpret an arbitrary
# entity in an Expression.


ID = _Field("id")
"""Get the "id" attribute from the entity."""



# Generated from Gram.g4 by ANTLR 4.10.1
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .GramParser import GramParser
else:
    from GramParser import GramParser


# This class defines a complete listener for a parse tree produced by GramParser.
class GramListener(ParseTreeListener):
    # Enter a parse tree produced by GramParser#parens.
    def enterParens(self, ctx: GramParser.ParensContext):
        pass

    # Exit a parse tree produced by GramParser#parens.
    def exitParens(self, ctx: GramParser.ParensContext):
        pass

    # Enter a parse tree produced by GramParser#compare.
    def enterCompare(self, ctx: GramParser.CompareContext):
        pass

    # Exit a parse tree produced by GramParser#compare.
    def exitCompare(self, ctx: GramParser.CompareContext):
        pass

    # Enter a parse tree produced by GramParser#logical.
    def enterLogical(self, ctx: GramParser.LogicalContext):
        pass

    # Exit a parse tree produced by GramParser#logical.
    def exitLogical(self, ctx: GramParser.LogicalContext):
        pass

    # Enter a parse tree produced by GramParser#literal.
    def enterLiteral(self, ctx: GramParser.LiteralContext):
        pass

    # Exit a parse tree produced by GramParser#literal.
    def exitLiteral(self, ctx: GramParser.LiteralContext):
        pass

    # Enter a parse tree produced by GramParser#array.
    def enterArray(self, ctx: GramParser.ArrayContext):
        pass

    # Exit a parse tree produced by GramParser#array.
    def exitArray(self, ctx: GramParser.ArrayContext):
        pass

    # Enter a parse tree produced by GramParser#function.
    def enterFunction(self, ctx: GramParser.FunctionContext):
        pass

    # Exit a parse tree produced by GramParser#function.
    def exitFunction(self, ctx: GramParser.FunctionContext):
        pass

    # Enter a parse tree produced by GramParser#attribute.
    def enterAttribute(self, ctx: GramParser.AttributeContext):
        pass

    # Exit a parse tree produced by GramParser#attribute.
    def exitAttribute(self, ctx: GramParser.AttributeContext):
        pass

    # Enter a parse tree produced by GramParser#boolean.
    def enterBoolean(self, ctx: GramParser.BooleanContext):
        pass

    # Exit a parse tree produced by GramParser#boolean.
    def exitBoolean(self, ctx: GramParser.BooleanContext):
        pass

    # Enter a parse tree produced by GramParser#string.
    def enterString(self, ctx: GramParser.StringContext):
        pass

    # Exit a parse tree produced by GramParser#string.
    def exitString(self, ctx: GramParser.StringContext):
        pass

    # Enter a parse tree produced by GramParser#int.
    def enterInt(self, ctx: GramParser.IntContext):
        pass

    # Exit a parse tree produced by GramParser#int.
    def exitInt(self, ctx: GramParser.IntContext):
        pass

    # Enter a parse tree produced by GramParser#float.
    def enterFloat(self, ctx: GramParser.FloatContext):
        pass

    # Exit a parse tree produced by GramParser#float.
    def exitFloat(self, ctx: GramParser.FloatContext):
        pass

    # Enter a parse tree produced by GramParser#null.
    def enterNull(self, ctx: GramParser.NullContext):
        pass

    # Exit a parse tree produced by GramParser#null.
    def exitNull(self, ctx: GramParser.NullContext):
        pass

    # Enter a parse tree produced by GramParser#args.
    def enterArgs(self, ctx: GramParser.ArgsContext):
        pass

    # Exit a parse tree produced by GramParser#args.
    def exitArgs(self, ctx: GramParser.ArgsContext):
        pass

    # Enter a parse tree produced by GramParser#nested_attr.
    def enterNested_attr(self, ctx: GramParser.Nested_attrContext):
        pass

    # Exit a parse tree produced by GramParser#nested_attr.
    def exitNested_attr(self, ctx: GramParser.Nested_attrContext):
        pass

    # Enter a parse tree produced by GramParser#attr.
    def enterAttr(self, ctx: GramParser.AttrContext):
        pass

    # Exit a parse tree produced by GramParser#attr.
    def exitAttr(self, ctx: GramParser.AttrContext):
        pass


del GramParser



import unittest
import tempfile
import pytest

import responses
from responses import matchers
import alligater.parse as parse
from crocodsl.field import _Field
from crocodsl.func import Hash, TrimPrefix

from .arm import Arm
from .feature import Feature
from .population import Population
from .rollout import Rollout
from .variant import Variant

FIXTURES = {
    # The simplest possible gate specification.
    "simple": {
        "feature": Feature(
            "simplest_feature",
            variants=[Variant("foo", "This is the only value")],
            default_arm=Arm("foo"),
        ),
        "yaml": """
        feature:
          name: simplest_feature
          variants:
            foo: This is the only value
          # Automatically roll out `foo` to everyone.
          default_arm: foo
        """,
    },
    # Simple A/B gate
    "simple_ab": {
        "feature": Feature(
            "ab_feature",
            variants=[
                Variant("a", "This is the value for 'a'"),
                Variant("b", "This is the value for 'b'"),
            ],
            rollouts=[
                Rollout(arms=[Arm("a"), Arm("b")]),
            ],
        ),
        "yaml": """
        feature:
          name: ab_feature
          variants:
            a: This is the value for 'a'
            b: This is the value for 'b'
          rollouts:
            # Automatically rollout to everyone with a 50/50 split.
            - arms:
              - a
              - b
        """,
    },
    # Fully specified A/B gate with unequal weights.
    "full_ab": {
        "feature": Feature(
            "ab_feature_full",
            variants=[
                Variant("a", "This is the value for 'a'"),
                Variant("b", "This is the value for 'b'"),
            ],
            rollouts=[
                Rollout(
                    name="default",
                    population=Population.DEFAULT,
                    arms=[
                        Arm("a", 0.25),
                        Arm("b", 0.75),
                    ],
                ),
            ],
        ),
        "yaml": """
        feature:
          name: ab_feature_full
          variants:
            a: This is the value for 'a'
            b: This is the value for 'b'
          rollouts:
            - name: default
              population: default
              arms:
                - variant: a
                  weight: 0.25
                - variant: b
                  weight: 0.75
        """,
    },
    # A/B with custom rollout randomizer
    "full_ab_custom": {
        "feature": Feature(
            "ab_feature_custom_randomizer",
            variants=[
                Variant("a", "This is the value for 'a'"),
                Variant("b", "This is the value for 'b'"),
            ],
            rollouts=[
                Rollout(
                    name="default",
                    population=Population.DEFAULT,
                    arms=[
                        Arm("a", 0.25),
                        Arm("b", 0.75),
                    ],
                    randomizer=Hash(_Field("custom")),
                ),
            ],
        ),
        "yaml": """
        feature:
          name: ab_feature_custom_randomizer
          variants:
            a: This is the value for 'a'
            b: This is the value for 'b'
          rollouts:
            - name: default
              population: default
              arms:
                - variant: a
                  weight: 0.25
                - variant: b
                  weight: 0.75
              randomizer: Hash($custom)
        """,
    },
    # Gate with multiple rollouts and populations. The second rollout is implied
    # from the `default_arm` setting.
    "multi_rollout": {
        "feature": Feature(
            "multi_rollout_feature",
            variants=[
                Variant("on", True),
                Variant("off", False),
            ],
            default_arm="off",
            rollouts=[
                Rollout(
                    name="test_segment",
                    population=Population.Percent(0.2, "some_seed"),
                    arms=["on"],
                ),
            ],
        ),
        "yaml": """
        feature:
          name: multi_rollout_feature
          variants:
            "on": true
            "off": false
          default_arm: "off"
          rollouts:
            - name: test_segment
              population:
                type: percent
                value: 0.2
                seed: some_seed
              arms:
                - "on"
        """,
    },
    # Fully-specified gate with multiple populations.
    "full_rollout": {
        "feature": Feature(
            name="multi_rollout_feature_full",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="temp_assignment_segment",
                    population=Population.Explicit(["id_temp"]),
                    arms=[Arm("off", weight=1.0)],
                    sticky=False,
                ),
                Rollout(
                    name="test_segment_1",
                    population=Population.Percent(0.2, "some_seed"),
                    arms=[
                        Arm("a", weight=0.5),
                        Arm("b", weight=0.5),
                    ],
                ),
                Rollout(
                    name="test_segment_2",
                    population=Population.Explicit(["id_1", "id_2", "id_3", "id_4"]),
                    arms=[Arm("a", weight=1.0)],
                ),
                Rollout(
                    name="test_segment_3",
                    population=Population.Percent(
                        0.5, "third_seed", id_field=_Field("non_standard_id")
                    ),
                    arms=["b", "off"],
                ),
            ],
        ),
        "yaml": """
        feature:
          name: multi_rollout_feature_full
          variants:
            a: 'A'
            b: 'B'
            "off": null
          default_arm: "off"
          rollouts:
            - name: temp_assignment_segment
              sticky: false
              population:
                type: explicit
                value: [id_temp]
              arms:
                - variant: "off"
                  weight: 1.0
            - name: test_segment_1
              population:
                type: percent
                value: 0.2
                seed: some_seed
              arms:
                - variant: a
                  weight: 0.5
                - variant: b
                  weight: 0.5
            - name: test_segment_2
              population:
                type: explicit
                value:
                  - id_1
                  - id_2
                  - id_3
                  - id_4
              arms:
                - variant: a
                  weight: 1.0
            - name: test_segment_3
              population:
                type: percent
                value: 0.5
                seed: third_seed
                field: $non_standard_id
              arms:
                - b
                - "off"
        """,
    },
    # Gate with custom field in explicit population
    "custom_explicit": {
        "feature": Feature(
            name="custom_explicit",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    population=Population.Explicit(
                        ["id_1", "id_2", "id_3", "id_4"], id_field=_Field("custom")
                    ),
                    arms=[Arm("a", weight=1.0)],
                ),
            ],
        ),
        "yaml": """
        feature:
          name: custom_explicit
          variants:
            a: 'A'
            b: 'B'
            "off": null
          default_arm: "off"
          rollouts:
            - name: test_segment_1
              population:
                type: explicit
                value:
                  - id_1
                  - id_2
                  - id_3
                  - id_4
                field: $custom
              arms:
                - variant: a
                  weight: 1.0
        """,
    },
    # Gate with custom expression in the explicit population field
    "custom_explicit_expression": {
        "feature": Feature(
            name="custom_explicit_expression",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    population=Population.Explicit(
                        ["id_1", "id_2", "id_3", "id_4"],
                        id_field=TrimPrefix(_Field("custom"), "prefix_"),
                    ),
                    arms=[Arm("a", weight=1.0)],
                ),
            ],
        ),
        "yaml": """
        feature:
          name: custom_explicit_expression
          variants:
            a: 'A'
            b: 'B'
            "off": null
          default_arm: "off"
          rollouts:
            - name: test_segment_1
              population:
                type: explicit
                value:
                  - id_1
                  - id_2
                  - id_3
                  - id_4
                field: TrimPrefix($custom, "prefix_")
              arms:
                - variant: a
                  weight: 1.0
        """,
    },
    # Gate with a custom expression in the population
    "custom_expression": {
        "feature": Feature(
            name="custom_expression",
            variants=[
                Variant("a", "A"),
                Variant("b", "B"),
                Variant("off", None),
            ],
            default_arm=Arm("off"),
            rollouts=[
                Rollout(
                    name="test_segment_1",
                    population=Population.Expression(
                        _Field("id").in_(["id_1", "id_2", "id_3", "id_4"]),
                    ),
                    arms=[Arm("a", weight=1.0)],
                ),
            ],
        ),
        "yaml": """
        feature:
          name: custom_expression
          variants:
            a: 'A'
            b: 'B'
            "off": null
          default_arm: "off"
          rollouts:
            - name: test_segment_1
              population:
                type: expression
                value: $id In ["id_1", "id_2", "id_3", "id_4"]
              arms:
                - variant: a
                  weight: 1.0
        """,
    },
}


class TestParse(unittest.TestCase):
    def assert_feature(self, name, objs):
        parsed = parse.parse_yaml(objs["yaml"])
        expected = objs["feature"]
        actual = parsed[expected.name]
        assert expected == actual, "Expected:\n{}\n\nActual:\n{}\n".format(
            expected.to_dict(), actual.to_dict()
        )

    def test_parse(self):
        for k, v in FIXTURES.items():
            self.assert_feature(k, v)

    def test_merge(self):
        parsed = parse.parse_yaml(
            FIXTURES["simple"]["yaml"],
            default_features={
                "simplest_feature": Feature(
                    "simplest_feature",
                    variants=[Variant("bar", "This will be overridden")],
                    default_arm=Arm("bar"),
                ),
            },
        )

        actual = parsed["simplest_feature"]
        expected = FIXTURES["simple"]["feature"]
        assert expected == actual, "Expected:\n{}\n\nActual:\n{}\n".format(
            expected.to_dict(), actual.to_dict()
        )

    def test_load_config_local(self):
        with tempfile.NamedTemporaryFile("w") as f:
            f.write("teststring")
            f.flush()
            parsed = parse.load_config(f.name)
            assert parsed == "teststring"

    @responses.activate
    def test_load_config_remote(self):
        responses.add(responses.GET, "http://example.com", body="teststring")
        parsed = parse.load_config("http://example.com")
        assert parsed == "teststring"

    @responses.activate
    def test_load_config_remote_auth(self):
        responses.add(
            responses.GET,
            "http://example.com",
            body="teststring",
            match=[matchers.header_matcher({"authorization": "Bearer token"})],
        )
        parsed = parse.load_config("http://example.com", authorization="Bearer token")
        assert parsed == "teststring"

    @responses.activate
    def test_load_config_remote_invalid(self):
        with pytest.raises(Exception) as exc_info:
            responses.add(responses.GET, "http://example.com", status=400)
            parse.load_config("http://example.com")
        assert str(exc_info.value) == "Failed to load config. Got status 400"

    def test_load_config_fn(self):
        parsed = parse.load_config(lambda: "foo")
        assert parsed == "foo"

    def test_load_config_fn_kwargs(self):
        def _loader(*, arg: str) -> str:
            return f"test: {arg}"

        parsed = parse.load_config(_loader, arg="foo")
        assert parsed == "test: foo"
